"""Transformer Decoder Stack."""

from typing import Dict, Optional, Sequence

import jax
import jax.numpy as jnp
import numpy as np

from ..core import Module
from ..nn import LayerNorm, Linear, MultiHeadAttention, RngSeq
from ..nn.dropout import dropout


class CausalSelfAttention(MultiHeadAttention):
    """Self attention with a causal mask applied."""

    def __call__(
        self,
        query: jnp.ndarray,
        key: Optional[jnp.ndarray] = None,
        value: Optional[jnp.ndarray] = None,
        mask: Optional[jnp.ndarray] = None,
    ) -> jnp.ndarray:
        key = key if key is not None else query
        value = value if value is not None else query

        seq_len = query.shape[1]
        causal_mask = np.tril(np.ones((seq_len, seq_len)))
        mask = mask * causal_mask if mask is not None else causal_mask

        return super().__call__(query, key, value, mask)


class DenseBlock(Module):
    """A 2-layer MLP which widens then narrows the input."""

    def __init__(self, in_dim: int, init_scale: float, widening_factor: int = 4):
        super().__init__()
        self._init_scale = init_scale
        initializer = jax.nn.initializers.variance_scaling(
            self._init_scale, mode="fan_in", distribution="normal"
        )
        self._widening_factor = widening_factor
        self.fc1 = Linear(in_dim, in_dim * widening_factor, w_init=initializer)
        self.fc2 = Linear(in_dim * widening_factor, in_dim, w_init=initializer)

    def __call__(self, x: jnp.ndarray) -> jnp.ndarray:
        x = self.fc1(x)
        x = jax.nn.gelu(x)
        return self.fc2(x)


class Transformer(Module):
    """A transformer stack."""

    layers: Sequence[Dict[str, Module]]

    def __init__(self, dim: int, num_heads: int, num_layers: int, dropout_rate: float):
        super().__init__()
        assert dim % num_heads == 0
        self._num_layers = num_layers
        self._num_heads = num_heads
        self._dropout_rate = dropout_rate

        self.rng_seq = RngSeq()

        init_scale = 2.0 / self._num_layers
        layers = []
        for _ in range(num_layers):
            layers.append(
                {
                    "attention": CausalSelfAttention(
                        num_heads=self._num_heads,
                        key_size=dim // num_heads,
                        w_init_scale=init_scale,
                    ),
                    "attn_layer_norm": LayerNorm(dim, -1, True, True),
                    "dense_layer_norm": LayerNorm(dim, -1, True, True),
                    "dense_block": DenseBlock(dim, init_scale),
                }
            )
        self.layers = layers
        self.layer_norm_output = LayerNorm(dim, -1, True, True)

    def __call__(
        self, h: jnp.ndarray, mask: Optional[jnp.ndarray] = None
    ) -> jnp.ndarray:
        """Connects the transformer.
        Args:
          h: Inputs, [B, T, H].
          mask: Padding mask, [B, T].
          is_training: Whether we're training or not.
        Returns:
          Array of shape [B, T, H].
        """

        dropout_rate = self._dropout_rate if self.training else 0.0
        if mask is not None:
            mask = mask[:, None, None, :]

        # Note: names chosen to approximately match those used in the GPT-2 code;
        # see https://github.com/openai/gpt-2/blob/master/src/model.py.
        rngs = self.rng_seq.next_rng_key(self._num_layers * 2)
        for i in range(self._num_layers):
            h_norm = self.layers[i]["attn_layer_norm"](h)
            h_attn = self.layers[i]["attention"](h_norm, mask=mask)
            h_attn = dropout(rngs[i * 2 + 0], dropout_rate, h_attn)
            h = h + h_attn
            h_norm = self.layers[i]["dense_layer_norm"](h)
            h_dense = self.layers[i]["dense_block"](h_norm)
            h_dense = dropout(rngs[i * 2 + 1], dropout_rate, h_dense)
            h = h + h_dense
        h = self.layer_norm_output(h)

        return h



from functools import partial

import jax
import jax.numpy as jnp
import pax


class UpsamplingNetwork(pax.Module):
    def __init__(self, n_mels, num_output_channels):
        super().__init__()
        self.input_conv = pax.nn.Conv1D(n_mels, 512, 3, padding="VALID", with_bias=True)

        dilated_conv = partial(pax.nn.Conv1D, 512, 512, 2, padding="VALID")
        self.dilated_conv_1 = dilated_conv(rate=1)
        self.dilated_conv_2 = dilated_conv(rate=2)
        self.dilated_conv_3 = dilated_conv(rate=4)

        conv1d_transpose = partial(pax.nn.Conv1DTranspose, 512, 512, padding="SAME")
        self.upsample_conv_1 = conv1d_transpose(kernel_shape=4, stride=4)
        self.upsample_conv_2 = conv1d_transpose(kernel_shape=2, stride=2)
        self.upsample_conv_3 = conv1d_transpose(kernel_shape=2, stride=2)
        self.output_conv = pax.nn.Conv1D(512, num_output_channels, 1, padding="VALID")

    def __call__(self, mel):
        x = self.input_conv(mel)
        res_1 = jax.nn.relu(self.dilated_conv_1(x))
        x = x[:, 1:] + res_1
        res_2 = jax.nn.relu(self.dilated_conv_2(x))
        x = x[:, 2:] + res_2
        res_3 = jax.nn.relu(self.dilated_conv_3(x))
        x = x[:, 4:] + res_3

        x = jax.nn.relu(self.upsample_conv_1(x))
        x = jax.nn.relu(self.upsample_conv_2(x))
        x = jax.nn.relu(self.upsample_conv_3(x))

        x = self.output_conv(x)

        # tile x16
        N, L, D = x.shape
        x = jnp.tile(x[:, :, None, :], (1, 1, 16, 1))
        x = jnp.reshape(x, (N, -1, D))

        return x


class WaveGRU(pax.Module):
    def __init__(self, n_mels, hidden_dim, n_mu_bits=8):
        super().__init__()
        self.n_mu_bits = n_mu_bits
        self.hidden_dim = hidden_dim

        self.upsampling = UpsamplingNetwork(n_mels, hidden_dim)
        self.gru = pax.nn.GRU(hidden_dim, hidden_dim)
        self.logits = pax.nn.Linear(hidden_dim, 2 ** n_mu_bits)
        self.embed = pax.nn.Embed(2 ** n_mu_bits, hidden_dim)

    def __call__(self, inputs):
        logmel, wav = inputs
        x = self.upsampling(logmel)
        hx = self.gru.initial_state(x.shape[0])
        wav = self.embed(wav)
        assert x.shape == wav.shape
        x = x + wav
        _, x = pax.scan(self.gru, hx, x, time_major=False)
        x = self.logits(x)
        return x

    def inference(self, logmel, rng_key=None):
        if rng_key is None:
            rng_key = pax.next_rng_key()

        x = jnp.array([2 ** (self.n_mu_bits - 1)], dtype=jnp.int32)
        hx = self.gru.initial_state(1)

        conds = self.upsampling(logmel)

        def loop(prev_state, inputs):
            x, hx, rng_key = prev_state
            rng_key, next_rng_key = jax.random.split(rng_key)

            x = self.embed(x) + inputs
            hx, x = self.gru(hx, x)
            x = self.logits(x)
            x = jax.random.categorical(rng_key, x)
            return (x, hx, next_rng_key), x

        _, x = pax.scan(loop, (x, hx, rng_key), conds, time_major=False)
        return x



# This file is an adaptation from
# https://raw.githubusercontent.com/deepmind/dm-haiku/main/docs/conf.py
# which is under Apache License, Version 2.0.


# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import doctest
import inspect
import os
import sys

sys.path.insert(0, os.path.abspath(".."))

import pax
import sphinxcontrib.katex as katex

# -- Project information -----------------------------------------------------

project = "PAX"
copyright = "2021, Thông Nguyễn"
author = "Thông Nguyễn"


# -- General configuration ---------------------------------------------------
master_doc = "index"


# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.doctest",
    "sphinx.ext.inheritance_diagram",
    "sphinx.ext.intersphinx",
    "sphinx.ext.linkcode",
    "sphinx.ext.napoleon",
    "sphinxcontrib.bibtex",
    "sphinxcontrib.katex",
    "sphinx_autodoc_typehints",
    "nbsphinx",
    "IPython.sphinxext.ipython_console_highlighting",
]


# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for autodoc -----------------------------------------------------

autodoc_default_options = {
    "member-order": "bysource",
    "special-members": True,
    "exclude-members": "__repr__, __str__, __weakref__",
}


# -- Options for HTML output -------------------------------------------------


# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ["_static"]


# -- Options for doctest -----------------------------------------------------

doctest_test_doctest_blocks = "true"
doctest_global_setup = """
import jax
import jax.numpy as jnp
import pax
import opax
pax.seed_rng_key(42)
"""
doctest_default_flags = (
    doctest.ELLIPSIS
    | doctest.IGNORE_EXCEPTION_DETAIL
    | doctest.DONT_ACCEPT_TRUE_FOR_1
    | doctest.NORMALIZE_WHITESPACE
)


# -- Options for katex ------------------------------------------------------

# See: https://sphinxcontrib-katex.readthedocs.io/en/0.4.1/macros.html
latex_macros = r"""
    \def \d              #1{\operatorname{#1}}
"""

# Translate LaTeX macros to KaTeX and add to options for HTML builder
katex_macros = katex.latex_defs_to_katex_macros(latex_macros)
katex_options = "macros: {" + katex_macros + "}"

# Add LaTeX macros for LATEX builder
latex_elements = {"preamble": latex_macros}


# -- Source code links -------------------------------------------------------


def linkcode_resolve(domain, info):
    """Resolve a GitHub URL corresponding to Python object."""
    if domain != "py":
        return None

    try:
        mod = sys.modules[info["module"]]
    except ImportError:
        return None

    obj = mod
    try:
        for attr in info["fullname"].split("."):
            obj = getattr(obj, attr)
    except AttributeError:
        return None
    else:
        obj = inspect.unwrap(obj)

    try:
        filename = inspect.getsourcefile(obj)
    except TypeError:
        return None

    try:
        source, lineno = inspect.getsourcelines(obj)
    except OSError:
        return None

    return "https://github.com/ntt123/pax/blob/main/pax/%s#L%d#L%d" % (
        os.path.relpath(filename, start=os.path.dirname(pax.__file__)),
        lineno,
        lineno + len(source) - 1,
    )


# -- nbsphinx configuration --------------------------------------------------

nbsphinx_execute = "never"
nbsphinx_codecell_lexer = "ipython"
nbsphinx_kernel_name = "python"
nbsphinx_timeout = 180
nbsphinx_prolog = r"""
{% set docname = 'docs/' + env.doc2path(env.docname, base=None) %}

.. only:: html

    .. role:: raw-html(raw)
        :format: html

    .. nbinfo::

        Interactive online version:
        :raw-html:`<a href="https://colab.research.google.com/github/ntt123/pax/blob/main/{{ docname }}"><img alt="Open In Colab" src="https://colab.research.google.com/assets/colab-badge.svg" style="vertical-align:text-bottom"></a>`
"""



"""
Manage thread local states
"""

import logging
import threading
import weakref
from contextlib import contextmanager
from typing import Any, Optional, Tuple, Union

import jax
import jax.numpy as jnp
import jax.tree_util

KeyArray = Union[Any, jnp.ndarray]


class PaxThreadingLocalState(threading.local):
    """Manage all thread local states used by PAX"""

    __slots__ = [
        "_enable_deep_copy",
        "_mutable_module_ref_list",
        "_rng_seed",
        "_rng_key",
    ]
    _enable_deep_copy: bool
    _mutable_module_ref_list: Tuple[weakref.ReferenceType, ...]
    _mutable_module_level: jax.core.Sublevel
    _rng_seed: Optional[int]
    _rng_key: Optional[KeyArray]

    def __init__(self):
        super().__init__()

        self._enable_deep_copy = False
        self._mutable_module_ref_list = ()
        self._mutable_module_level = jax.core.cur_sublevel()
        self._rng_seed = None
        self._rng_key = None

    def is_deep_copy_enabled(self):
        """use deepcopy to copy modules"""
        return self._enable_deep_copy

    def add_mutable_module(self, module):
        """add `module` to mutable list"""
        self._mutable_module_ref_list = (
            weakref.ref(module),
            *self._mutable_module_ref_list,
        )

    def is_mutable(self, module):
        """Is `module` mutable?"""

        # cannot modify a module whose level of abstraction
        # is lower than the current level
        if self._mutable_module_level < jax.core.cur_sublevel():
            return False

        for ref in self._mutable_module_ref_list:
            if module is ref():
                return True

        return False

    @contextmanager
    def enable_deep_copy(self):
        r"""A context manager that turns on deepcopy mode."""
        prev = self._enable_deep_copy
        self._enable_deep_copy = True
        try:
            yield
        finally:
            self._enable_deep_copy = prev

    @contextmanager
    def allow_mutation(self, modules):
        r"""A context manager that turns on mutability."""
        if not isinstance(modules, (tuple, list)):
            modules = (modules,)
        modules = tuple(weakref.ref(mod) for mod in modules)

        prev = self._mutable_module_ref_list
        prev_abstraction_level = self._mutable_module_level
        prev_rng_state = self.get_rng_state()
        try:
            self._mutable_module_ref_list = modules
            self._mutable_module_level = jax.core.cur_sublevel()
            yield
        finally:
            self._mutable_module_ref_list = prev
            self._mutable_module_level = prev_abstraction_level
            self.set_rng_state(prev_rng_state)

    def seed_rng_key(self, seed: int) -> None:
        """Set ``self.rng_seed = seed``.
        Reset ``self.rng_key`` to ``None``.

        Arguments:
            seed: an integer seed.
        """
        assert isinstance(seed, int)
        self._rng_seed = seed
        self._rng_key = None

    def next_rng_key(self) -> KeyArray:
        """Return a random rng key.
        Renew the global random key ``self.rng_key``.

        If ``self.rng_key`` is ``None``,
        generate a new ``self.rng_key`` from ``self.rng_seed``.
        """
        if self._rng_key is None:
            if self._rng_seed is None:
                seed = 42
                logging.warning(
                    "Seeding RNG key with seed %s. "
                    "Use `pax.seed_rng_key` function to avoid this warning.",
                    seed,
                )
                self.seed_rng_key(seed)

            # Delay the generating of self.rng_key until `next_rng_key` is called.
            # This helps to avoid the problem when `seed_rng_key` is called
            # before jax found TPU cores.
            if self._rng_seed is not None:
                self._rng_key = jax.random.PRNGKey(self._rng_seed)
            else:
                raise ValueError("Impossible")

        key, self._rng_key = jax.random.split(self._rng_key)

        return key

    def get_rng_state(self):
        """Return internal states."""
        return (self._rng_key, self._rng_seed)

    def set_rng_state(self, state):
        """Set internal states."""
        rng_key, seed = state
        self._rng_key = rng_key
        self._rng_seed = seed


PAX_STATE = PaxThreadingLocalState()
add_mutable_module = PAX_STATE.add_mutable_module
allow_mutation = PAX_STATE.allow_mutation
enable_deep_copy = PAX_STATE.enable_deep_copy
get_rng_state = PAX_STATE.get_rng_state
is_deep_copy_enabled = PAX_STATE.is_deep_copy_enabled
is_mutable = PAX_STATE.is_mutable
next_rng_key = PAX_STATE.next_rng_key
seed_rng_key = PAX_STATE.seed_rng_key
set_rng_state = PAX_STATE.set_rng_state



# Source:
# https://github.com/lucidrains/denoising-diffusion-pytorch/blob/master/denoising_diffusion_pytorch/denoising_diffusion_pytorch.py

import math
from functools import partial
from inspect import isfunction
from typing import List

import jax
import jax.numpy as jnp
import numpy as np
import pax
from einops import rearrange
from pax.nn import GroupNorm, LayerNorm


def exists(x):
    return x is not None


def default(val, d):
    if exists(val):
        return val
    return d() if isfunction(d) else d


class Residual(pax.Module):
    def __init__(self, fn):
        super().__init__()
        self.fn = fn

    def __call__(self, x, *args, **kwargs):
        return self.fn(x, *args, **kwargs) + x


class SinusoidalPosEmbed(pax.Module):
    def __init__(self, dim):
        super().__init__()
        self.dim = dim

    def __call__(self, x):
        half_dim = self.dim // 2
        emb = math.log(10_000) / (half_dim - 1)
        emb = jnp.exp(jnp.arange(0, half_dim) * (-emb))
        emb = x[:, None] * emb[None, :]
        emb = jnp.concatenate((jnp.sin(emb), jnp.cos(emb)), axis=-1)
        return emb


class Mish(pax.Module):
    def __call__(self, x):
        return x * jnp.tanh(jax.nn.softplus(x))


class Upsample(pax.Module):
    def __init__(self, dim):
        super().__init__()
        self.conv = pax.nn.Conv2DTranspose(dim, dim, 4, 2, padding="SAME")

    def __call__(self, x):
        return self.conv(x)


class Downsample(pax.Module):
    def __init__(self, dim):
        super().__init__()
        self.conv = pax.nn.Conv2D(dim, dim, 3, 2, padding="SAME")

    def __call__(self, x):
        return self.conv(x)


class PreNorm(pax.Module):
    def __init__(self, dim, fn):
        super().__init__()
        self.fn = fn
        self.norm = LayerNorm(dim, -1, True, True)

    def __call__(self, x):
        x = self.norm(x)
        return self.fn(x)


class Block(pax.Module):
    def __init__(self, dim, dim_out, groups: int = 8):
        super().__init__()
        self.blocks = pax.nn.Sequential(
            pax.nn.Conv2D(dim, dim_out, 3, padding="SAME"),
            GroupNorm(groups, dim_out),
            Mish(),
        )

    def __call__(self, x):
        return self.blocks(x)


class ResnetBlock(pax.Module):
    def __init__(self, dim, dim_out, *, time_emb_dim=None, groups=8):
        super().__init__()
        self.mlp = (
            pax.nn.Sequential(Mish(), pax.nn.Linear(time_emb_dim, dim_out))
            if exists(time_emb_dim)
            else None
        )
        self.block1 = Block(dim, dim_out, groups=groups)
        self.block2 = Block(dim_out, dim_out, groups=groups)

        if dim != dim_out:
            self.res_conv = pax.nn.Conv2D(dim, dim_out, 1)
        else:
            self.res_conv = pax.nn.Identity()

    def __call__(self, x, time_emb):
        h = self.block1(x)

        if exists(self.mlp):
            h = h + self.mlp(time_emb)[:, None, None, :]

        h = self.block2(h)
        return h + self.res_conv(x)


class LinearAttention(pax.Module):
    def __init__(self, dim, heads=4, dim_head=32):
        super().__init__()
        self.heads = heads
        hidden_dim = dim_head * heads
        self.to_qkv = pax.nn.Conv2D(dim, hidden_dim * 3, 1, with_bias=False)
        self.to_out = pax.nn.Conv2D(hidden_dim, dim, 1)

    def __call__(self, x):
        b, h, w, c = x.shape
        qkv = self.to_qkv(x)
        q, k, v = rearrange(
            qkv, "b h w (qkv heads c) -> qkv b heads c (h w)", heads=self.heads, qkv=3
        )
        k = jax.nn.softmax(k, axis=-1)
        context = jnp.einsum("bhdn,bhen->bhde", k, v)
        out = jnp.einsum("bhde,bhdn->bhen", context, q)
        out = rearrange(
            out, "b heads c (h w) -> b h w (heads c)", heads=self.heads, h=h, w=w
        )
        return self.to_out(out)


### model


class UNet(pax.Module):
    ups: List[List[pax.Module]]
    downs: List[List[pax.Module]]

    def __init__(
        self,
        dim,
        out_dim=None,
        dim_mults=(1, 2, 4, 8),
        groups=8,
        channels=3,
        with_time_emb=True,
    ):
        super().__init__()
        self.channels = channels

        dims = [channels] + [dim * m for m in dim_mults]
        in_out = list(zip(dims[:-1], dims[1:]))

        if with_time_emb:
            time_dim = dim
            self.time_mlp = pax.nn.Sequential(
                SinusoidalPosEmbed(dim),
                pax.nn.Linear(dim, dim * 4),
                Mish(),
                pax.nn.Linear(dim * 4, dim),
            )
        else:
            time_dim = None
            self.time_mlp = None

        self.downs = []
        self.ups = []
        num_resolutions = len(in_out)

        for ind, (dim_in, dim_out) in enumerate(in_out):
            is_last = ind >= (num_resolutions - 1)
            self.downs.append(
                [
                    ResnetBlock(dim_in, dim_out, time_emb_dim=time_dim, groups=groups),
                    ResnetBlock(dim_out, dim_out, time_emb_dim=time_dim, groups=groups),
                    Residual(PreNorm(dim_out, LinearAttention(dim_out))),
                    Downsample(dim_out) if not is_last else pax.nn.Identity(),
                ]
            )

        mid_dim = dims[-1]
        self.mid_block1 = ResnetBlock(
            mid_dim, mid_dim, time_emb_dim=time_dim, groups=groups
        )
        self.mid_attn = Residual(PreNorm(mid_dim, LinearAttention(mid_dim)))
        self.mid_block2 = ResnetBlock(
            mid_dim, mid_dim, time_emb_dim=time_dim, groups=groups
        )

        for ind, (dim_in, dim_out) in enumerate(reversed(in_out[1:])):
            is_last = ind >= (num_resolutions - 1)

            self.ups.append(
                [
                    ResnetBlock(
                        dim_out * 2, dim_in, time_emb_dim=time_dim, groups=groups
                    ),
                    ResnetBlock(dim_in, dim_in, time_emb_dim=time_dim, groups=groups),
                    Residual(PreNorm(dim_in, LinearAttention(dim_in))),
                    Upsample(dim_in) if not is_last else pax.nn.Identity(),
                ]
            )

        out_dim = default(out_dim, channels)

        self.final_conv = pax.nn.Sequential(
            Block(dim, dim, groups=groups),
            pax.nn.Conv2D(dim, out_dim, 1),
        )

    def __call__(self, x, time):
        t = self.time_mlp(time) if exists(self.time_mlp) else None
        h = []
        for resnet, resnet2, attn, downsample in self.downs:
            x = resnet(x, t)
            x = resnet2(x, t)
            x = attn(x)
            h.append(x)
            x = downsample(x)

        x = self.mid_block1(x, t)
        x = self.mid_attn(x)
        x = self.mid_block2(x, t)

        for resnet, resnet2, attn, upsample in self.ups:
            x = jnp.concatenate((x, h.pop()), axis=-1)
            x = resnet(x, t)
            x = resnet2(x, t)
            x = attn(x)
            x = upsample(x)

        return self.final_conv(x)


def extract(a, t, x_shape):
    b, *_ = t.shape
    out = jnp.take_along_axis(a, t, axis=-1)
    out = jnp.reshape(out, (b,) + (1,) * (len(x_shape) - 1))
    return out


def noise_like(rng_key, shape, repeat=False):
    repeat_noise = lambda: jnp.tile(
        jax.random.normal(rng_key, (1, *shape[1:])),
        (shape[0],) + (1,) * (len(shape) - 1),
    )
    noise = lambda: jax.random.normal(rng_key, shape)
    return repeat_noise() if repeat else noise()


def cosine_beta_schedule(timesteps, s=0.008):
    """
    cosine schedule
    as proposed in https://openreview.net/forum?id=-NEXDKk8gZ
    """
    steps = timesteps + 1
    x = np.linspace(0, steps, steps)
    alphas_cumprod = np.cos(((x / steps) + s) / (1 + s) * np.pi * 0.5) ** 2
    alphas_cumprod = alphas_cumprod / alphas_cumprod[0]
    betas = 1 - (alphas_cumprod[1:] / alphas_cumprod[:-1])
    return np.clip(betas, a_min=0, a_max=0.999)


class GaussianDiffusion(pax.Module):
    def __init__(
        self,
        denoise_fn,
        *,
        image_size,
        channels=3,
        timesteps=1000,
        loss_type="l1",
        betas=None,
        random_seed=42,
    ):
        super().__init__()
        self.channels = channels
        self.image_size = image_size
        self.denoise_fn = denoise_fn
        self.rng_seq = pax.nn.RngSeq(random_seed)

        if exists(betas):
            betas = jnp.array(betas)
        else:
            betas = cosine_beta_schedule(timesteps)

        alphas = 1.0 - betas
        alphas_cumprod = np.cumprod(alphas, axis=0)
        alphas_cumprod_prev = np.append(1.0, alphas_cumprod[:-1])
        (timesteps,) = betas.shape
        self.num_timesteps = int(timesteps)
        self.loss_type = loss_type

        with self.add_states():
            self.betas = betas
            self.alphas_cumprod = alphas_cumprod
            self.alphas_cumprod_prev = alphas_cumprod_prev

            # calculations for diffusion q(x_t | x_{t-1}) and others
            self.sqrt_alphas_cumprod = np.sqrt(alphas_cumprod)
            self.sqrt_one_minus_alphas_cumprod = np.sqrt(1.0 - alphas_cumprod)
            self.log_one_minus_alphas_cumprod = np.log(1.0 - alphas_cumprod)
            self.sqrt_recip_alphas_cumprod = np.sqrt(1.0 / alphas_cumprod)
            self.sqrt_recipm1_alphas_cumprod = np.sqrt(1.0 / alphas_cumprod - 1)

            # calculations for posterior q(x_{t-1} | x_t, x_0)
            posterior_variance = (
                betas * (1.0 - alphas_cumprod_prev) / (1.0 - alphas_cumprod)
            )
            # above: equal to 1. / (1. / (1. - alpha_cumprod_tm1) + alpha_t / beta_t)
            self.posterior_variance = posterior_variance
            # below: log calculation clipped because the posterior variance is 0
            # at the beginning of the diffusion chain
            self.posterior_log_variance_clipped = np.log(
                np.maximum(posterior_variance, 1e-20)
            )
            self.posterior_mean_coef1 = (
                betas * np.sqrt(alphas_cumprod_prev) / (1.0 - alphas_cumprod)
            )
            self.posterior_mean_coef2 = (
                (1.0 - alphas_cumprod_prev) * np.sqrt(alphas) / (1.0 - alphas_cumprod)
            )

    def q_mean_variance(self, x_start, t):
        mean = extract(self.sqrt_alphas_cumprod, t, x_start.shape) * x_start
        variance = extract(1.0 - self.alphas_cumprod, t, x_start.shape)
        log_variance = extract(self.log_one_minus_alphas_cumprod, t, x_start.shape)
        return mean, variance, log_variance

    def predict_start_from_noise(self, x_t, t, noise):
        return (
            extract(self.sqrt_recip_alphas_cumprod, t, x_t.shape) * x_t
            - extract(self.sqrt_recipm1_alphas_cumprod, t, x_t.shape) * noise
        )

    def q_posterior(self, x_start, x_t, t):
        posterior_mean = (
            extract(self.posterior_mean_coef1, t, x_t.shape) * x_start
            + extract(self.posterior_mean_coef2, t, x_t.shape) * x_t
        )
        posterior_variance = extract(self.posterior_variance, t, x_t.shape)
        posterior_log_variance_clipped = extract(
            self.posterior_log_variance_clipped, t, x_t.shape
        )
        return posterior_mean, posterior_variance, posterior_log_variance_clipped

    def p_mean_variance(self, x, t, clip_denoised: bool):
        x_recon = self.predict_start_from_noise(x, t=t, noise=self.denoise_fn(x, t))

        if clip_denoised:
            x_recon = jnp.clip(x_recon, -1.0, 1.0)

        model_mean, posterior_variance, posterior_log_variance = self.q_posterior(
            x_start=x_recon, x_t=x, t=t
        )
        return model_mean, posterior_variance, posterior_log_variance

    def p_sample(self, x, t, rng_key, clip_denoised=True, repeat_noise=False):
        b = x.shape[0]
        model_mean, _, model_log_variance = self.p_mean_variance(
            x=x, t=t, clip_denoised=clip_denoised
        )
        noise = noise_like(rng_key, x.shape, repeat_noise)
        # no noise when t == 0
        nonzero_mask = jnp.reshape(
            (1 - (t == 0).astype(jnp.float32)),
            (b,) + (1,) * (len(x.shape) - 1),
        )

        return model_mean + nonzero_mask * jnp.exp(0.5 * model_log_variance) * noise

    def p_sample_loop(self, shape, rng_key):
        b = shape[0]
        rng_key_, rng_key = jax.random.split(rng_key)
        img = jax.random.normal(rng_key_, shape)

        i_s = jnp.flip(jnp.arange(0, self.num_timesteps, dtype=jnp.int32))
        rng_keys = jax.random.split(rng_key, self.num_timesteps)

        def loop_fn(img, inputs):
            i, rng_key = inputs
            img = self.p_sample(img, jnp.full((b,), i, dtype=jnp.int32), rng_key)
            return img, None

        img, _ = pax.scan(loop_fn, img, (i_s, rng_keys))
        return img

    @partial(jax.jit, static_argnums=[1, 2])
    def sample(self, batch_size=16, random_seed=42):
        image_size = self.image_size
        channels = self.channels
        return self.p_sample_loop(
            (batch_size, image_size, image_size, channels),
            jax.random.PRNGKey(random_seed),
        )

    def q_sample(self, x_start, t, noise=None):
        noise = default(
            noise,
            lambda: jax.random.normal(
                self.rng_seq.next_rng_key(), x_start.shape, x_start.dtype
            ),
        )

        return (
            extract(self.sqrt_alphas_cumprod, t, x_start.shape) * x_start
            + extract(self.sqrt_one_minus_alphas_cumprod, t, x_start.shape) * noise
        )

    def p_losses(self, x_start, t, noise=None):
        noise = default(
            noise,
            lambda: jax.random.normal(
                self.rng_seq.next_rng_key(), x_start.shape, x_start.dtype
            ),
        )

        x_noisy = self.q_sample(x_start=x_start, t=t, noise=noise)
        x_recon = self.denoise_fn(x_noisy, t)

        if self.loss_type == "l1":
            loss = jnp.mean(jnp.abs(noise - x_recon))
        elif self.loss_type == "l2":
            loss = jnp.mean(jnp.square(noise - x_recon))
        else:
            raise NotImplementedError()

        return loss

    def __call__(self, x, *args, **kwargs):
        b, h, w, c = x.shape
        img_size = self.image_size
        assert (
            h == img_size and w == img_size
        ), f"height and width of image must be {img_size}"

        t = jax.random.randint(
            self.rng_seq.next_rng_key(),
            (b,),
            minval=0,
            maxval=self.num_timesteps,
            dtype=jnp.int32,
        )
        return self.p_losses(x, t, *args, **kwargs)



"""Enforce mixed-precision policy."""

import functools
from typing import TypeVar

import jax.numpy as jnp
import jmp

from .module import Module, PaxKind

T = TypeVar("T", bound=Module)

# source: https://stackoverflow.com/a/21963090
def _find_descriptor(cls, attrname):
    """Find the descriptor of an attribute."""

    def hasspecialmethod(obj, name):
        return any(name in klass.__dict__ for klass in type(obj).__mro__)

    for klass in cls.__mro__:
        if attrname in klass.__dict__:
            descriptor = klass.__dict__[attrname]
            if not hasspecialmethod(descriptor, "__get__"):
                return None
            return descriptor
    return None


def _wrap_method(func):
    """Wrap a class's method to enforce mixe-precision policy."""

    @functools.wraps(func)
    def mp_method_wrapper(self, *args, **kwargs):
        """A mixed-precision method.

        - Convert all weights to compute dtype.
        - Cast all arguments to compute dtype.
        - Call the original method.
        - Convert all weights to param dtype.
        - Cast output to output dtype.

        We bypass PAX mutability checking to make mixed-precision
        policy transparent from the user's point of view.
        """
        original_values = {}
        casted_original = {}
        # pylint: disable=protected-access

        # convert weights to compute dtype
        for name, kind in self._pax.name_to_kind.items():
            if kind in [PaxKind.PARAMETER, PaxKind.STATE]:
                value = getattr(self, name)
                casted_value = self._pax_mp_policy.cast_to_compute(value)
                self.__dict__[name] = casted_value
                original_values[name] = value
                casted_original[name] = casted_value

        # cast arguments to compute dtype
        args, kwargs = self._pax_mp_policy.cast_to_compute((args, kwargs))
        output = func.__get__(self, type(self))(*args, **kwargs)  # type:ignore

        # convert weights to param dtype
        for name, kind in self._pax.name_to_kind.items():
            if kind in [PaxKind.PARAMETER, PaxKind.STATE]:
                value = getattr(self, name)
                if value is not casted_original[name]:  # modified
                    casted_value = self._pax_mp_policy.cast_to_param(value)
                    setattr(self, name, casted_value)
                else:
                    # avoid casting operation
                    self.__dict__[name] = original_values[name]

        # cast output to output dtype
        output = self._pax_mp_policy.cast_to_output(output)
        return output

    return mp_method_wrapper


def _mp_repr(mp_policy):
    dtype_to_name = {
        jnp.bfloat16: "H",
        jnp.float16: "H",
        jnp.float32: "F",
        jnp.float64: "F",
    }

    return (
        dtype_to_name[mp_policy.param_dtype]
        + dtype_to_name[mp_policy.compute_dtype]
        + dtype_to_name[mp_policy.output_dtype]
    )


def apply_mp_policy(module: T, mp_policy: jmp.Policy) -> T:
    """Create a mixed-precision module.

    Create a subclass on the fly to enforce the mixed-precision policy.

    >>> import jmp
    >>> mp_policy = jmp.get_policy("params=float32,compute=float16,output=float32")
    >>> net = pax.nn.Linear(3, 3)
    >>> net = pax.apply_mp_policy(net, mp_policy)
    >>> print(net.summary())
    Linear(in_dim=3, out_dim=3, with_bias=True, mp_policy=FHF)
    """

    if hasattr(module, "_pax_mp_policy"):
        raise ValueError(
            "Cannot apply multiple mixed-precision policies on an object.\n"
            "Call `pax.unwrap_mp_policy(...)` to remove the policy first."
        )

    # pylint: disable=protected-access
    cls_name = module.__class__.__name__
    module_methods = dir(Module)
    base = module.__class__

    methods = {}
    for name in dir(base):
        if name != "__call__" and name.startswith("__"):
            continue
        if name == "__call__" or name not in module_methods:
            value = getattr(base, name)
            if callable(value):
                value = _find_descriptor(base, name)
                if value is None:
                    continue
                if isinstance(value, (staticmethod, classmethod)):
                    methods[name] = value
                else:
                    methods[name] = _wrap_method(value)

    def _repr(self, info=None):
        if info is None:
            info = {}
        info["mp_policy"] = _mp_repr(self._pax_mp_policy)
        return super(base, self)._repr(info)  # type: ignore

    methods["_repr"] = _repr

    cls = type(cls_name, (base,), methods)
    obj = object.__new__(cls)
    object.__setattr__(obj, "_pax", module._pax)
    obj.__dict__.update(module.__dict__)
    obj.__dict__["_pax_mp_policy"] = mp_policy
    for name, kind in obj._pax.name_to_kind.items():
        if kind in [PaxKind.PARAMETER, PaxKind.STATE]:
            obj.__dict__[name] = mp_policy.cast_to_param(obj.__dict__[name])
    return obj


def unwrap_mp_policy(module: T) -> T:
    """Unwrap a mixed-precision module to recreate the original module.

    >>> import jmp
    >>> mp_policy = jmp.get_policy("params=float32,compute=float16,output=float32")
    >>> net = pax.nn.Linear(3, 3)
    >>> net = pax.apply_mp_policy(net, mp_policy)
    >>> print(net.summary())
    Linear(in_dim=3, out_dim=3, with_bias=True, mp_policy=FHF)
    >>> net = pax.unwrap_mp_policy(net)
    >>> print(net.summary())
    Linear(in_dim=3, out_dim=3, with_bias=True)
    """
    if not hasattr(module, "_pax_mp_policy"):
        raise ValueError("Expected a mixed-precision module.")

    base = module.__class__.__base__
    original = object.__new__(base)
    object.__setattr__(original, "_pax", module._pax)
    original.__dict__.update(module.__dict__)
    del original.__dict__["_pax_mp_policy"]
    return original



import math

import fire
import jax
import jax.numpy as jnp
import numpy as np
import opax
import pax
import tensorflow as tf
from PIL import Image
from tqdm.auto import tqdm

from data_loader import load_celeb_a
from model import GaussianDiffusion, UNet


def make_image_grid(images, padding=2):
    """Place images in a square grid."""
    n = images.shape[0]
    size = int(math.sqrt(n))
    assert size * size == n, "expecting a square grid"
    img = images[0]

    H = img.shape[0] * size + padding * (size + 1)
    W = img.shape[1] * size + padding * (size + 1)
    out = np.zeros((H, W, img.shape[-1]), dtype=img.dtype)
    for i in range(n):
        x = i % size
        y = i // size
        xstart = x * (img.shape[0] + padding) + padding
        xend = xstart + img.shape[0]
        ystart = y * (img.shape[1] + padding) + padding
        yend = ystart + img.shape[1]
        out[xstart:xend, ystart:yend, :] = images[i]
    return out


def train(
    batch_size: int = 32,
    learning_rate: float = 1e-4,
    num_training_steps: int = 10_000,
    log_freq: int = 1000,
    image_size: int = 64,
    random_seed: int = 42,
):

    pax.seed_rng_key(random_seed)

    model = UNet(dim=64, dim_mults=(1, 2, 4, 8))

    diffusion = GaussianDiffusion(
        model,
        image_size=image_size,
        timesteps=1000,
        loss_type="l1",  # L1 or L2
    )

    dataset = load_celeb_a()

    dataloader = (
        dataset.repeat()
        .shuffle(batch_size * 100)
        .batch(batch_size)
        .take(num_training_steps)
        .prefetch(tf.data.AUTOTUNE)
    )

    def loss_fn(model, inputs):
        model, loss = pax.module_and_value(model)(inputs)
        return loss, (loss, model)

    update_fn = pax.utils.build_update_fn(loss_fn)
    fast_update_fn = jax.jit(update_fn)

    optimizer = opax.adam(learning_rate)(diffusion.parameters())

    total_loss = 0.0
    tr = tqdm(dataloader)
    for step, batch in enumerate(tr, 1):
        batch = jax.tree_map(lambda x: x.numpy(), batch)
        diffusion, optimizer, loss = fast_update_fn(diffusion, optimizer, batch)
        total_loss = total_loss + loss

        if step % log_freq == 0:
            loss = total_loss / log_freq
            total_loss = 0.0
            tr.write(f"[step {step:05d}]  train loss {loss:.3f}")

            imgs = jax.device_get(diffusion.eval().sample(16))
            imgs = ((imgs * 0.5 + 0.5) * 255).astype(jnp.uint8)
            imgs = make_image_grid(imgs)
            im = Image.fromarray(imgs)
            im.save(f"sample_{step:05d}.png")


if __name__ == "__main__":
    fire.Fire(train)



# Follow resnet implementation at:
# https://github.com/deepmind/dm-haiku/blob/main/haiku/_src/nets/resnet.py
# which is under Apache License, Version 2.0.

"""Resnet Modules."""

from typing import Sequence, Tuple

import jax
import jax.numpy as jnp

from ..core import Module
from ..nn import BatchNorm2D, Conv2D, Linear, max_pool


class ResnetBlock(Module):
    """ResnetBlock"""

    layers: Sequence[Tuple[Conv2D, BatchNorm2D]]

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        stride,
        use_projection: bool,
        bottleneck: bool,
    ):
        super().__init__()
        self.use_projection = use_projection

        if self.use_projection:
            self.proj_conv = Conv2D(
                in_channels,
                out_channels,
                kernel_shape=1,
                stride=stride,
                with_bias=False,
                padding=[(0, 0), (0, 0)],
                data_format="NCHW",
                name="proj_conv",
            )
            self.proj_batchnorm = BatchNorm2D(
                out_channels, True, True, 0.9, data_format="NCHW", name="proj_bn"
            )

        channel_div = 4 if bottleneck else 1
        conv_0 = Conv2D(
            in_features=in_channels,
            out_features=out_channels // channel_div,
            kernel_shape=1 if bottleneck else 3,
            stride=1 if bottleneck else stride,
            with_bias=False,
            padding=[(0, 0), (0, 0)] if bottleneck else [(1, 1), (1, 1)],
            data_format="NCHW",
            name="conv1",
        )

        bn_0 = BatchNorm2D(
            out_channels // channel_div, True, True, 0.9, data_format="NCHW", name="bn1"
        )

        conv_1 = Conv2D(
            in_features=out_channels // channel_div,
            out_features=out_channels,
            kernel_shape=3,
            stride=stride if bottleneck else 1,
            with_bias=False,
            padding=[(1, 1), (1, 1)],
            data_format="NCHW",
            name="conv2",
        )

        bn_1 = BatchNorm2D(
            out_channels, True, True, 0.9, data_format="NCHW", name="bn2"
        )

        layers = ((conv_0, bn_0), (conv_1, bn_1))

        if bottleneck:
            conv_2 = Conv2D(
                in_features=out_channels,
                out_features=out_channels,
                kernel_shape=1,
                stride=1,
                with_bias=False,
                padding=[(0, 0), (0, 0)],
                data_format="NCHW",
                name="conv3",
            )
            bn_2 = BatchNorm2D(
                out_channels,
                True,
                True,
                0.9,
                data_format="NCHW",
                name="bn3",
            )
            layers = layers + ((conv_2, bn_2),)

        self.layers = layers

    def __call__(self, inputs):
        out = shortcut = inputs

        if self.use_projection:
            shortcut = self.proj_conv(shortcut)
            shortcut = self.proj_batchnorm(shortcut)

        for i, (conv_i, bn_i) in enumerate(self.layers):
            out = conv_i(out)
            out = bn_i(out)

            if i < len(self.layers) - 1:
                out = jax.nn.relu(out)

        return jax.nn.relu(out + shortcut)


class BlockGroup(Module):
    """Group of Blocks."""

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        num_blocks: int,
        stride,
        bottleneck: bool,
        use_projection,
    ):
        super().__init__()

        blocks = []

        for i in range(num_blocks):
            blocks.append(
                ResnetBlock(
                    in_channels=(in_channels if i == 0 else out_channels),
                    out_channels=out_channels,
                    stride=(1 if i else stride),
                    use_projection=(i == 0 and use_projection),
                    bottleneck=bottleneck,
                )
            )

        self.blocks = blocks

    def __call__(self, inputs):
        out = inputs
        for block in self.blocks:
            out = block(out)

        return out


def check_length(length, value, name):
    if len(value) != length:
        raise ValueError(f"`{name}` must be of length 4 not {len(value)}")


class ResNet(Module):
    """A generic ResNet module."""

    CONFIGS = {
        18: {
            "blocks_per_group": (2, 2, 2, 2),
            "bottleneck": False,
            "channels_per_group": (64, 128, 256, 512),
            "use_projection": (False, True, True, True),
        },
        34: {
            "blocks_per_group": (3, 4, 6, 3),
            "bottleneck": False,
            "channels_per_group": (64, 128, 256, 512),
            "use_projection": (False, True, True, True),
        },
        50: {
            "blocks_per_group": (3, 4, 6, 3),
            "bottleneck": True,
            "channels_per_group": (256, 512, 1024, 2048),
            "use_projection": (True, True, True, True),
        },
        101: {
            "blocks_per_group": (3, 4, 23, 3),
            "bottleneck": True,
            "channels_per_group": (256, 512, 1024, 2048),
            "use_projection": (True, True, True, True),
        },
        152: {
            "blocks_per_group": (3, 8, 36, 3),
            "bottleneck": True,
            "channels_per_group": (256, 512, 1024, 2048),
            "use_projection": (True, True, True, True),
        },
        200: {
            "blocks_per_group": (3, 24, 36, 3),
            "bottleneck": True,
            "channels_per_group": (256, 512, 1024, 2048),
            "use_projection": (True, True, True, True),
        },
    }

    def __init__(
        self,
        input_channels: int,
        blocks_per_group: Sequence[int],
        num_classes: int,
        bottleneck: bool = True,
        channels_per_group: Sequence[int] = (256, 512, 1024, 2048),
        use_projection: Sequence[bool] = (True, True, True, True),
        logits_config=None,
        initial_conv_config=None,
        name=None,
    ):
        super().__init__(name=name)

        check_length(4, blocks_per_group, "blocks_per_group")
        check_length(4, channels_per_group, "channels_per_group")

        logits_config = dict(logits_config or {})
        logits_config.setdefault("w_init", jax.nn.initializers.zeros)

        initial_conv_config = dict(initial_conv_config or {})
        initial_conv_config.setdefault("in_features", input_channels)
        initial_conv_config.setdefault("out_features", 64)
        initial_conv_config.setdefault("kernel_shape", 7)
        initial_conv_config.setdefault("stride", 2)
        initial_conv_config.setdefault("with_bias", False)
        initial_conv_config.setdefault("padding", [(3, 3), (3, 3)])
        initial_conv_config.setdefault("data_format", "NCHW")

        self.initial_conv = Conv2D(**initial_conv_config, name="conv1")

        self.initial_batchnorm = BatchNorm2D(
            initial_conv_config["out_features"],
            True,
            True,
            0.9,
            data_format="NCHW",
            name="bn1",
        )

        block_groups = []
        strides = (1, 2, 2, 2)
        for i in range(4):
            block_groups.append(
                BlockGroup(
                    in_channels=(
                        initial_conv_config["out_features"]
                        if i == 0
                        else channels_per_group[i - 1]
                    ),
                    out_channels=channels_per_group[i],
                    num_blocks=blocks_per_group[i],
                    stride=strides[i],
                    bottleneck=bottleneck,
                    use_projection=use_projection[i],
                )
            )

        self.block_groups = block_groups

        self.logits = Linear(
            channels_per_group[-1], num_classes, **logits_config, name="fc"
        )

    def __call__(self, inputs):
        out = inputs
        out = self.initial_conv(out)
        out = self.initial_batchnorm(out)
        out = jax.nn.relu(out)
        out = jnp.pad(out, [(0, 0), (0, 0), (1, 1), (1, 1)])
        out = max_pool(
            out,
            window_shape=(1, 1, 3, 3),
            strides=(1, 1, 2, 2),
            padding="VALID",
            channel_axis=1,
        )
        for block_group in self.block_groups:
            out = block_group(out)

        out = jnp.mean(out, axis=(2, 3))
        return self.logits(out)


class ResNet18(ResNet):
    """ResNet18."""

    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        logits_config=None,
        initial_conv_config=None,
    ):
        super().__init__(
            input_channels=input_channels,
            num_classes=num_classes,
            initial_conv_config=initial_conv_config,
            logits_config=logits_config,
            **ResNet.CONFIGS[18],
            name="ResNet18",
        )


class ResNet34(ResNet):
    """ResNet34."""

    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        logits_config=None,
        initial_conv_config=None,
    ):
        super().__init__(
            input_channels=input_channels,
            num_classes=num_classes,
            initial_conv_config=initial_conv_config,
            logits_config=logits_config,
            **ResNet.CONFIGS[34],
        )


class ResNet50(ResNet):
    """ResNet50."""

    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        logits_config=None,
        initial_conv_config=None,
    ):
        super().__init__(
            input_channels=input_channels,
            num_classes=num_classes,
            initial_conv_config=initial_conv_config,
            logits_config=logits_config,
            **ResNet.CONFIGS[50],
        )


class ResNet101(ResNet):
    """ResNet101."""

    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        logits_config=None,
        initial_conv_config=None,
    ):
        super().__init__(
            input_channels=input_channels,
            num_classes=num_classes,
            initial_conv_config=initial_conv_config,
            logits_config=logits_config,
            **ResNet.CONFIGS[101],
        )


class ResNet152(ResNet):
    """ResNet152."""

    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        logits_config=None,
        initial_conv_config=None,
    ):
        super().__init__(
            input_channels=input_channels,
            num_classes=num_classes,
            initial_conv_config=initial_conv_config,
            logits_config=logits_config,
            **ResNet.CONFIGS[152],
        )


class ResNet200(ResNet):
    """ResNet200."""

    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        logits_config=None,
        initial_conv_config=None,
    ):
        super().__init__(
            input_channels=input_channels,
            num_classes=num_classes,
            initial_conv_config=initial_conv_config,
            logits_config=logits_config,
            **ResNet.CONFIGS[200],
        )



"""Useful functions."""

from typing import TypeVar
from unittest import TestCase

import jax

from .base import BaseModule

T = TypeVar("T", bound=BaseModule)


def assert_structure_equal(tree_a: T, tree_b: T):
    """Assert that the two pytrees are structurally the same.

    Print out the difference.
    """
    if jax.tree_structure(tree_a) == jax.tree_structure(tree_b):
        return True

    def check(subtree_a, subtree_b):
        if isinstance(subtree_a, BaseModule) and isinstance(subtree_b, BaseModule):
            assert_structure_equal(subtree_a, subtree_b)

    has_error = False
    try:
        jax.tree_map(
            check,
            tree_a,
            tree_b,
            is_leaf=lambda x: isinstance(x, BaseModule)
            and x is not tree_a
            and x is not tree_b,
        )
    except ValueError:
        has_error = True

    if has_error:
        test_case = TestCase()
        test_case.maxDiff = None
        # do not compare weights
        tree_a_w_none_leaves = jax.tree_map(lambda _: None, tree_a)
        tree_b_w_none_leaves = jax.tree_map(lambda _: None, tree_b)
        test_case.assertDictEqual(
            vars(tree_a_w_none_leaves), vars(tree_b_w_none_leaves)
        )

    return has_error



"""Transform a module to a new one."""
from collections import OrderedDict
from types import MappingProxyType
from typing import Any, TypeVar

import jax

from .base import BaseModule, EmptyNode, PaxKind

TreeDef = Any

T = TypeVar("T", bound=BaseModule)
K = TypeVar("K", bound=BaseModule)
O = TypeVar("O", bound=BaseModule)


def _update_pax(mod, pax_info):
    super(BaseModule, mod).__setattr__("_pax", pax_info)
    return mod


def enable_train_mode(mod: T) -> T:
    """Return a module in training mode."""

    def _train_apply_fn(mod: T) -> T:
        # pylint: disable=protected-access
        return _update_pax(mod, mod._pax._replace(training=True))

    return mod.apply(_train_apply_fn)


def enable_eval_mode(mod: T) -> T:
    """Return a module in evaluation mode."""

    def _eval_apply_fn(mod: T) -> T:
        # pylint: disable=protected-access
        return _update_pax(mod, mod._pax._replace(training=False))

    return mod.apply(_eval_apply_fn)


def freeze_parameters(mod: T) -> T:
    """Return a copy module with all trainable parameters are converted to non-trainable states."""

    def _freeze_apply_fn(mod: T) -> T:
        new_name_to_kind = OrderedDict()
        # pylint: disable=protected-access
        for (name, kind) in mod._pax.name_to_kind.items():
            if kind == PaxKind.PARAMETER:
                new_name_to_kind[name] = PaxKind.STATE
            else:
                new_name_to_kind[name] = kind

        # use proxy to avoid any side effects
        # pylint: disable=protected-access
        pax_info = mod._pax._replace(name_to_kind=MappingProxyType(new_name_to_kind))
        return _update_pax(mod, pax_info)

    return mod.apply(_freeze_apply_fn)


def unfreeze_parameters(mod: T, *, origin: T) -> T:
    """Return a copy module with all trainable parameters are converted to non-trainable states."""

    assert freeze_parameters(origin) == mod
    tree_def = jax.tree_structure(origin)
    leaves = jax.tree_leaves(mod)
    return jax.tree_unflatten(tree_def, leaves)


def select_kind(mod: T, *, kind: PaxKind) -> T:
    """Select leaves of kind ``kind`` while setting all other leaves to ``None``.

    Arguments:
        mod: The module.
        kind: The kind of leaves that will be kept intact.
    """
    assert kind in [PaxKind.PARAMETER, PaxKind.STATE]
    if kind == PaxKind.STATE:
        none_list = [PaxKind.PARAMETER]
    else:
        none_list = [PaxKind.STATE]

    def _select_apply_fn(mod: T) -> T:
        # pylint: disable=protected-access
        for (name, kind) in mod._pax.name_to_kind.items():
            if kind in none_list:
                value = getattr(mod, name)
                none_v = jax.tree_map(lambda _: EmptyNode(), value)
                mod.__dict__[name] = none_v
        return mod

    return mod.apply(_select_apply_fn)


def select_parameters(mod: T) -> T:
    """Select `PARAMETER` leaves only."""
    return select_kind(mod, kind=PaxKind.PARAMETER)


def select_states(mod: T) -> T:
    """Select `STATE` leaves only."""
    return select_kind(mod, kind=PaxKind.STATE)


def update_pytree(mod: T, *, other: T) -> T:
    """Use non-EmptyNode leaves from other."""

    def _select_fn(leaf_x, leaf_y):
        if isinstance(leaf_y, EmptyNode):
            return leaf_x
        else:
            return leaf_y

    is_empty = lambda x: isinstance(x, EmptyNode)
    new_mod = jax.tree_map(_select_fn, mod, other, is_leaf=is_empty)
    return new_mod


def update_parameters(mod: T, *, params: T) -> T:
    """Return a module that uses trainable parameters in `params`."""
    return update_pytree(mod, other=select_parameters(params))


def update_states(mod: T, *, states: T) -> T:
    """Return a module that uses non-trainable states in `states`."""
    return update_pytree(mod, other=select_states(states))



"""PAX mechanisms to make PAX functions pure."""

import functools
import gc
from types import MethodType
from typing import Optional, Sequence, Union

import jax

from .base import BaseModule
from .threading_local import (
    allow_mutation,
    enable_deep_copy,
    get_rng_state,
    set_rng_state,
)


def _get_modules(tree):
    "Return a list of modules in the pytree `tree`."
    modules = jax.tree_flatten(tree, is_leaf=lambda x: isinstance(x, BaseModule))[0]
    modules = [m for m in modules if isinstance(m, BaseModule)]
    return modules


def _get_all_submodules(value):
    submods = _get_modules(value)
    out = list(submods)
    for mod in submods:
        out.extend(_get_all_submodules(mod.submodules()))
    return out


def pure(
    func,
    static_argnums: Optional[Union[int, Sequence[int]]] = None,
    check_leaks: bool = True,
):
    """Make a function pure by copying the inputs.

    Any modification on the copy will not affect the original inputs.

    **Note**: only functions that are wrapped by `pax.pure` are allowed to modify PAX's Modules.

    Example:

    >>> f = pax.nn.Linear(3,3)
    >>> f.a_list = []
    Traceback (most recent call last):
      ...
    ValueError: Cannot modify a module in immutable mode.
    Please do this computation inside a function decorated by `pax.pure`.
    >>>
    >>> @pax.pure
    ... def add_list(m):
    ...     m.a_list = []
    ...     return m
    ...
    >>> f = add_list(f)
    >>> print(f.a_list)
    []

    Arguments:
        func: A function.
        static_argnums: a list of static arguments.
        check_leaks: enable jax leak checking.

    Returns:
        A pure function.
    """

    rng_state = get_rng_state()

    if isinstance(static_argnums, int):
        static_argnums = (static_argnums,)

    if static_argnums is None:
        static_argnums = ()

    def _deepcopy(value):
        with enable_deep_copy():
            leaves, treedef = jax.tree_flatten(value)
        return jax.tree_unflatten(treedef, leaves)

    @functools.wraps(func)
    def _f(*args, **kwargs):
        _ = [m.scan_bugs() for m in _get_modules((func, args, kwargs))]

        # support calling method
        if isinstance(func, MethodType):
            args = (func.__self__, *args)
            unbound_func = func.__func__
        # or calling a module
        elif isinstance(func, BaseModule) and callable(func):
            args = (func, *args)
            unbound_func = func.__call__.__func__
        elif callable(func):
            unbound_func = func
        else:
            raise ValueError("Not supported")

        args = list(args)
        args_copy = tuple(args)
        for i in static_argnums:
            args[i] = None
        args = tuple(args)

        def no_leak_func(*args, **kwargs):
            args, kwargs = _deepcopy((args, kwargs))
            modules = _get_all_submodules((args, kwargs))
            with allow_mutation(modules):
                gc.collect()
                args = list(args)
                for i in static_argnums:
                    args[i] = args_copy[i]
                args = tuple(args)
                set_rng_state(rng_state)
                out = unbound_func(*args, **kwargs)
                set_rng_state(rng_state)
                gc.collect()
                return out

        def _run(args, kwargs, eval_shape: bool = False):
            if eval_shape:
                out = jax.eval_shape(no_leak_func, *args, **kwargs)
            else:
                out = no_leak_func(*args, **kwargs)
            return out

        with jax.check_tracer_leaks(check_leaks):
            if check_leaks:
                _run(args, kwargs, eval_shape=True)
            # real run
            out = _run(args, kwargs, eval_shape=False)
        _ = [m.scan_bugs() for m in _get_modules(out)]
        return out

    return _f



import numpy as np
import pax
import torchvision

IMAGENET_MEAN = np.array((0.485, 0.456, 0.406))
IMAGENET_STD = np.array((0.229, 0.224, 0.225))


def convert_conv(conv, name=None):
    """Return a pax.nn.Conv2D module with weights from pretrained ``conv``."""
    weight = conv.weight.data.contiguous().permute(2, 3, 1, 0).contiguous().numpy()[:]

    pax_conv = pax.nn.Conv2D(
        in_features=conv.in_channels,
        out_features=conv.out_channels,
        kernel_shape=conv.kernel_size,
        stride=conv.stride,
        with_bias=False,
        padding=[(conv.padding[0],) * 2, (conv.padding[1],) * 2],
        data_format="NCHW",
        name=name,
    )
    assert pax_conv.weight.shape == weight.shape
    return pax_conv.replace(weight=weight)


def convert_bn(bn, name=None):
    """Return a pax.nn.BatchNorm2D module from pretrained ``bn``."""
    weight = bn.weight.data.numpy()[None, :, None, None]
    bias = bn.bias.data.numpy()[None, :, None, None]
    running_mean = bn.running_mean.data.numpy()[None, :, None, None]
    running_var = bn.running_var.data.numpy()[None, :, None, None]

    pax_bn = pax.nn.BatchNorm2D(
        num_channels=bias.shape[1],
        create_offset=True,
        create_scale=True,
        decay_rate=0.9,
        eps=1e-5,
        data_format="NCHW",
        name=name,
    )
    assert pax_bn.scale.shape == weight.shape
    assert pax_bn.offset.shape == bias.shape
    assert pax_bn.ema_mean.averages.shape == running_mean.shape
    assert pax_bn.ema_var.averages.shape == running_var.shape

    pax_bn = pax_bn.replace(scale=weight, offset=bias)
    pax_bn = pax_bn.replace_node(pax_bn.ema_mean.averages, running_mean)
    pax_bn = pax_bn.replace_node(pax_bn.ema_var.averages, running_var)
    return pax_bn


def convert_basic_block(block):
    conv1 = convert_conv(block.conv1, name="conv1")
    bn1 = convert_bn(block.bn1, name="bn1")
    conv2 = convert_conv(block.conv2, name="conv2")
    bn2 = convert_bn(block.bn2, name="bn2")

    if block.downsample is not None:
        conv0 = convert_conv(block.downsample[0], name="proj_conv")
        bn0 = convert_bn(block.downsample[1], name="proj_bn")
        return ((conv1, bn1), (conv2, bn2)), (conv0, bn0)
    else:
        return (((conv1, bn1), (conv2, bn2)),)


def convert_block_group(group):
    out = []
    for i in range(len(group)):
        out.append(convert_basic_block(group[i]))
    return out


def convert_linear(linear):
    weight = linear.weight.data.numpy()
    bias = linear.bias.data.numpy()
    pax_linear = pax.nn.Linear(
        in_dim=weight.shape[1], out_dim=weight.shape[0], with_bias=True
    )
    weight = np.transpose(weight)
    assert pax_linear.bias.shape == bias.shape
    assert pax_linear.weight.shape == weight.shape

    return pax_linear.replace(weight=weight, bias=bias)


def load_pretrained_resnet18():
    resnet18 = pax.nets.ResNet18(3, 1000)
    resnet18_pt = torchvision.models.resnet18(pretrained=True).eval()
    pax_resnet = [
        convert_conv(resnet18_pt.conv1),
        convert_bn(resnet18_pt.bn1),
        convert_block_group(resnet18_pt.layer1),
        convert_block_group(resnet18_pt.layer2),
        convert_block_group(resnet18_pt.layer3),
        convert_block_group(resnet18_pt.layer4),
        convert_linear(resnet18_pt.fc),
    ]

    def replace_parts(resnet18):
        # replace resnet18 part by part
        resnet18.initial_conv = pax_resnet[0]
        resnet18.initial_batchnorm = pax_resnet[1]
        for i in range(len(resnet18.block_groups)):
            bg = resnet18.block_groups[i]
            for j in range(len(bg.blocks)):
                b = bg.blocks[j]
                mods = pax_resnet[2 + i][j]
                b.layers = mods[0]
                if b.use_projection:
                    b.proj_conv = mods[1][0]
                    b.proj_batchnorm = mods[1][1]

        resnet18.logits = pax_resnet[-1]
        # make sure we are in `eval` mode when doing evaluation.
        return resnet18.eval()

    return pax.pure(replace_parts)(resnet18)



"""PAX mechanisms to make PAX method pure."""

from functools import partial
from types import MethodType
from typing import Callable, Optional, Sequence, Tuple, TypeVar, Union

from .base import BaseModule
from .pure import pure

O = TypeVar("O")
T = TypeVar("T", bound=BaseModule)


def module_and_value(
    module_or_method: Callable[..., O],
    static_argnums: Optional[Union[int, Sequence[int]]] = None,
    check_leaks: bool = True,
) -> Callable[..., Tuple[T, O]]:
    """Return a pure function that executes a module's method.

    This pure function also returns the updated input module in the output.

    Example:

    >>> net = pax.nn.Linear(1, 1)
    >>> x = jnp.ones((32, 1))
    >>> net, y = pax.module_and_value(net)(x)  # note: `net` is also returned.


    Arguments:
        module_or_method: Either a PAX module or a method of a PAX module.
        static_argnums: a list of static arguments.
        check_leaks: enable jax leak checking.

    Returns:
        A pure function.
    """
    if isinstance(module_or_method, MethodType):  # a method
        mod = module_or_method.__self__
        func = module_or_method.__func__
    elif isinstance(module_or_method, BaseModule):  # a module
        mod = module_or_method
        assert hasattr(mod, "__call__"), "Expecting a callable module."
        func = module_or_method.__call__.__func__
    else:
        raise ValueError("Expecting a module or a module's method.")

    assert isinstance(mod, BaseModule), "Expecting a PAX module."

    if static_argnums is not None:
        if isinstance(static_argnums, int):
            static_argnums = [static_argnums]
        # offset by 1 for `self` argument.
        static_argnums = tuple(x + 1 for x in static_argnums)

    @partial(pure, static_argnums=static_argnums, check_leaks=check_leaks)
    def _run(mod, *args, **kwargs):
        out = func(mod, *args, **kwargs)
        return mod, out

    return partial(_run, mod)



"""PAX BaseModule."""

# Note: This file is originated from
# https://raw.githubusercontent.com/cgarciae/treex/32e4cce5ca0cc991cda8076903853621d0aa4ab9/treex/module.py
# which is under MIT License.

import functools
from collections import OrderedDict
from contextlib import contextmanager
from copy import deepcopy
from enum import Enum
from types import MappingProxyType
from typing import (
    Any,
    Iterable,
    List,
    Mapping,
    NamedTuple,
    Optional,
    Tuple,
    Type,
    TypeVar,
)

import jax
import jax.numpy as jnp
import jax.tree_util
import numpy as np
from jax.dtypes import issubdtype as isdt

# pylint: disable=no-name-in-module
from jaxlib.xla_extension import CompiledFunction

from .threading_local import allow_mutation, is_deep_copy_enabled, is_mutable

T = TypeVar("T", bound="BaseModule")
M = TypeVar("M")


@jax.tree_util.register_pytree_node_class
class EmptyNode(Tuple):
    """We use this class to mark deleted nodes.

    Note: this is inspired by treex's `Nothing` class.
    """

    def tree_flatten(self):
        """Flatten empty node."""
        return (), None

    @classmethod
    def tree_unflatten(cls, aux, children):
        """Unflatten empty node."""
        del aux, children
        return EmptyNode()


class PaxKind(Enum):
    """``PaxKind`` lists all supported attribute kinds in ``pax.Module``.

    An attribute will be considered as part of the pytree structure
    if its kind is one of ``STATE``, ``PARAMETER``, ``MODULE``.

    * A ``STATE`` attribute is a non-trainable leaf of the pytree.
    * A ``PARAMETER`` attribute is a trainable leaf of the pytree.
    * A ``MODULE`` attribute is a generic subtree.
    """

    STATE: int = 1
    PARAMETER: int = 2
    MODULE: int = 3
    UNKNOWN: int = -1


class PaxModuleInfo(NamedTuple):
    """PAX Internal Data Structure."""

    training: bool
    name_to_kind: Mapping[str, PaxKind]
    default_kind: PaxKind

    def __repr__(self) -> str:
        nodes = ", ".join([f"{k}:{v.name}" for k, v in self.name_to_kind.items()])
        return f"PaxModuleInfo[training={self.training}, nodes={{{nodes}}}]"


class BaseModuleMetaclass(type):
    """Metaclass for `BaseModule`."""

    def __call__(cls: Type[T], *args, **kwargs) -> T:
        module = cls.__new__(cls, *args, **kwargs)  # type: ignore

        with allow_mutation(module):
            with module._default_kind(module._pax.default_kind):
                with module._default_kind(PaxKind.MODULE):
                    cls.__init__(module, *args, **kwargs)

        # scan module after initialization for potential bugs
        module._assert_not_shared_module()
        module._assert_not_shared_weight()
        module._scan_fields(module.__dict__.keys())
        return module


class BaseModule(metaclass=BaseModuleMetaclass):
    """BaseModule manages all information related to the pytree.

    There are two important methods:

    - ``tree_flatten`` converts a module to ``(leaves, treedef)``
    - ``tree_unflatten`` restores the module.

    BaseModule maintains a ``name_to_kind`` dictionary that tells if an attribute is part of
    the pytree and the kind of the tree part (parameter, state, module, etc.).
    """

    __slots__ = ("_pax",)

    _pax: PaxModuleInfo

    @property
    def pax(self):
        return self._pax

    def __new__(cls: Type[T], *args, **kwargs) -> T:
        """Initialize `_pax` attribute in `__new__` method to avoid
        calling `super().__init__()` in every subclass of Module."""

        del args, kwargs

        obj = object.__new__(cls)
        super(BaseModule, obj).__setattr__(
            "_pax",
            PaxModuleInfo(
                training=True,
                name_to_kind=MappingProxyType(OrderedDict()),
                default_kind=PaxKind.UNKNOWN,
            ),
        )

        if obj.__slots__ != ("_pax",):
            raise ValueError("`__slots__` is not supported by PAX modules.")

        # scan class attributes for unregistered modules and ndarrays.
        obj._scan_fields(obj._class_fields())

        return obj

    def _class_fields(self):
        for name, value in self.__class__.__dict__.items():
            if not hasattr(value, "__get__"):  # ignore descriptors
                yield name

    def _assert_mutability(self):
        if not is_mutable(self):
            raise ValueError(
                "Cannot modify a module in immutable mode.\n"
                "Please do this computation inside a function decorated by `pax.pure`."
            )

    def _update_name_to_kind_dict(self, name: str, value):
        """Update the `name_to_kind` dictionary.

        Create a new dictionary and wrap it with `MappingProxyType` to avoid side effects."""
        self._assert_mutability()

        new_dict = OrderedDict(self._pax.name_to_kind)
        new_dict[name] = value
        new_info = self._pax._replace(name_to_kind=MappingProxyType(new_dict))
        super().__setattr__("_pax", new_info)

    def __setattr__(self, name: str, value: Any) -> None:
        """Setting `_pax` attribute is forbidden."""
        self._assert_mutability()

        if name == "_pax":
            raise ValueError(
                f"{name} is a reserved attribute for PAX internal mechanisms."
            )

        super().__setattr__(name, value)

    def __delattr__(self, name: str) -> None:
        del name
        raise ValueError("Cannot delete a module's attribute.")

    def _update_default_kind(self, kind: PaxKind):
        self._assert_mutability()
        new_pax_info = self._pax._replace(default_kind=kind)
        super(BaseModule, self).__setattr__("_pax", new_pax_info)

    @contextmanager
    def _default_kind(self, kind: PaxKind):
        """Search for any new unregistered pytree attribute
        at the end of the context and register it with kind `kind`.
        """
        fields_before = set(vars(self).keys())
        try:
            yield
        finally:
            fields_after = vars(self).keys()
            new_fields = (name for name in fields_after if name not in fields_before)
            self._find_and_register_pytree(kind, fields=new_fields)

    def add_parameters(self):
        """Return a context manager that registers new attributes
        in the scope as trainable parameters.

        >>> def add_param(net):
        ...     with net.add_parameters():
        ...         net.weight = jnp.array(0.)
        ...         net.bias = jnp.array(0.)
        ...     return net
        ...
        >>> net = pax.pure(add_param)(pax.Module())
        >>> net._pax
        PaxModuleInfo[training=True, nodes={weight:PARAMETER, bias:PARAMETER}]
        """
        return self._default_kind(PaxKind.PARAMETER)

    def add_states(self):
        """Return a context manager that registers new attributes
        in the scope as non-trainable states.
        """
        return self._default_kind(PaxKind.STATE)

    def tree_flatten(
        self,
    ) -> Tuple[List[jnp.ndarray], Tuple[Mapping[str, Any], PaxModuleInfo]]:
        """Convert a module to ``(children, treedef)``."""

        aux = dict(self.__dict__)
        children = [aux.pop(name) for name in self._pax.name_to_kind]

        if is_deep_copy_enabled():
            leaves, treedef = jax.tree_flatten(aux)
            new_leaves = []
            black_list = (jax.custom_jvp, functools.partial, CompiledFunction)
            for leaf in leaves:
                try:
                    if isinstance(leaf, black_list):
                        new_leaves.append(leaf)
                    else:
                        new_leaf = deepcopy(leaf)
                        new_leaves.append(new_leaf)
                except TypeError:
                    new_leaves.append(leaf)
            aux = jax.tree_unflatten(treedef, new_leaves)

        return children, (aux, self._pax)

    @classmethod
    def tree_unflatten(cls, aux_pax, children):
        """Recreate a module from its ``(children, treedef)``."""
        aux, pax_info = aux_pax
        module = object.__new__(cls)
        super(BaseModule, module).__setattr__("_pax", pax_info)
        module_dict = module.__dict__
        module_dict.update(aux)
        # don't have to copy `name_to_kind` anymore, speed thing up!
        # md["name_to_kind"] = OrderedDict(module.name_to_kind)
        # pylint: disable=protected-access
        module_dict.update(zip(module._pax.name_to_kind, children))

        return module

    def __init_subclass__(cls):
        """Any subclass of ``Module`` is also registered as pytree."""
        jax.tree_util.register_pytree_node_class(cls)

    def _scan_fields(self, fields: Iterable[str]):
        """Scan fields for *potential* bugs."""

        for name in fields:
            value = getattr(self, name)
            kind = self._pax.name_to_kind.get(name, PaxKind.UNKNOWN)
            mods, _ = jax.tree_flatten(
                value, is_leaf=lambda x: isinstance(x, BaseModule)
            )
            leaves = jax.tree_leaves(value)

            # Check if a MODULE attribute contains non-module leafs.
            if kind == PaxKind.MODULE:
                for mod in mods:
                    if not isinstance(mod, BaseModule):
                        raise ValueError(
                            f"Field `{self}.{name}` (kind={kind}, value={value}) "
                            f"contains a non-module leaf (type={type(mod)}, value={mod})."
                        )

            # Check if a pytree attribute contains non-ndarray values.
            if kind != PaxKind.UNKNOWN:
                for leaf in leaves:
                    if not isinstance(leaf, (np.ndarray, jnp.ndarray)):
                        raise ValueError(
                            f"Field `{self}.{name}` ({kind}) contains a non-ndarray value "
                            f"(type={type(leaf)}, value={leaf})."
                        )

            # Check if a PARAMETER attribute contains non-differentiable ndarrays.
            if kind == PaxKind.PARAMETER:
                for leaf in leaves:
                    if hasattr(leaf, "dtype") and not (
                        isdt(leaf.dtype, jnp.complexfloating)
                        or isdt(leaf.dtype, jnp.floating)
                    ):
                        raise ValueError(
                            f"Field `{self}.{name}` ({kind}) contains a non-differentiable leaf "
                            f"(type={leaf.dtype}, value={leaf})."
                        )

            if kind == PaxKind.UNKNOWN:
                # if a field contains empty pytree
                leaves, _ = jax.tree_flatten(
                    value, is_leaf=lambda x: isinstance(x, BaseModule)
                )

                # Check if a field contains unregistered module
                for leaf in mods:
                    if isinstance(leaf, BaseModule):
                        raise ValueError(
                            f"Unregistered field `{self}.{name}` ({kind}) contains a Module "
                            f"({leaf}). "
                            f"Consider registering it as a MODULE."
                        )

                # Check if a field contains unregistered ndarray
                for leaf in leaves:
                    if isinstance(leaf, (np.ndarray, jnp.ndarray)):
                        raise ValueError(
                            f"Unregistered field `{self.__class__.__name__}.{name}`"
                            f" ({kind}) contains a ndarray. Consider registering it using"
                            f" `self.add_parameters()` or `self.add_states()` context managers."
                        )

            # Check if an unregistered (or PARAMETER) field contains pax.Module instances
            if kind not in [PaxKind.MODULE, PaxKind.STATE]:
                for mod in mods:
                    if isinstance(mod, BaseModule):
                        raise ValueError(
                            f"Field `{self}.{name}` ({kind}) "
                            f"SHOULD NOT contains a pax.Module instance {mod}."
                        )

    def __eq__(self, o: object) -> bool:
        """Compare two modules."""
        if id(self) == id(o):
            return True

        if type(self) is not type(o):
            return False

        self_leaves, self_treedef = jax.tree_flatten(self)
        o_leaves, o_treedef = jax.tree_flatten(o)

        if len(self_leaves) != len(o_leaves):
            return False

        if self_treedef != o_treedef:
            return False

        leaves_equal = jax.tree_map(lambda a, b: a is b, self_leaves, o_leaves)
        return all(leaves_equal)

    def __hash__(self) -> int:
        leaves, treedef = jax.tree_flatten(self)
        leaves = jax.tree_map(lambda x: (x.shape, x.dtype), leaves)
        return hash((tuple(leaves), treedef))

    def _find_and_register_pytree(
        self, kind: PaxKind, fields: Optional[Iterable[str]] = None
    ):
        """Find unregistered pytree attributes and register it with kind `kind`."""

        if kind == PaxKind.UNKNOWN:
            return

        if fields is not None:
            items = ((name, getattr(self, name)) for name in fields)
        else:
            items = vars(self).items()

        def is_pytree_of(tree, classes):
            is_leaf_fn = lambda m: isinstance(m, classes)
            leaves = jax.tree_flatten(tree, is_leaf=is_leaf_fn)[0]
            return len(leaves) > 0 and all(is_leaf_fn(m) for m in leaves)

        for name, value in items:
            if name not in self._pax.name_to_kind:
                if (kind == PaxKind.MODULE and is_pytree_of(value, BaseModule)) or (
                    kind in [PaxKind.PARAMETER, PaxKind.STATE]
                    and not _has_module_node(value)
                    and is_pytree_of(value, (np.ndarray, jnp.ndarray))
                ):
                    self._update_name_to_kind_dict(name, kind)

    def register_subtree(self, name: str, value, kind: PaxKind):
        """Assign `value` to attribute `name` and register its kind as `kind`."""
        if hasattr(self, name):
            raise RuntimeError("Cannot register an existing attribute")

        setattr(self, name, value)
        self.set_attribute_kind(**{name: kind})

    def set_attribute_kind(self, **kwargs):
        """Set PAX kind for one or many attributes.

        Example:

        >>> class M(pax.Module):
        ...     def __init__(self):
        ...         super().__init__()
        ...         self.weight = jnp.array(0.)
        ...         self.counter = jnp.array(0)
        ...         self.set_attribute_kind(weight=pax.P, counter=pax.S)
        ...
        >>> m = M()
        >>> print(m._pax.name_to_kind)
        OrderedDict([('weight', <PaxKind.PARAMETER: 2>), ('counter', <PaxKind.STATE: 1>)])
        """
        for name, kind in kwargs.items():
            if not hasattr(self, name):
                raise AttributeError(f"Attribute `{name}` does not exist.")
            self._update_name_to_kind_dict(name, kind)

    def apply(self: T, apply_fn) -> T:
        """Apply a function to all submodules.

        >>> def print_param_count(mod):
        ...     count = sum(jax.tree_leaves(jax.tree_map(jnp.size, mod)))
        ...     print(f"{count}\t{mod}")
        ...     return mod
        ...
        >>> net = pax.nn.Sequential(pax.nn.Linear(1, 1), jax.nn.relu)
        >>> net = net.apply(print_param_count)
        2 Linear(in_dim=1, out_dim=1, with_bias=True)
        0 Lambda(relu)
        2 Sequential

        Arguments:
            apply_fn: a function which inputs a module and outputs a transformed module.
            check_treedef: check treedef before applying the function.
        """

        def rec_fn(mod_or_ndarray):
            if isinstance(mod_or_ndarray, BaseModule):
                return mod_or_ndarray.apply(apply_fn)
            else:
                return mod_or_ndarray

        submodules: List[BaseModule] = self.submodules()
        new_self = jax.tree_map(
            rec_fn,
            self,
            is_leaf=lambda x: isinstance(x, BaseModule) and (x in submodules),
        )

        # tree_map already created a copy of self,
        # hence `apply_fn` is guaranteed to have no side effects.
        return apply_fn(new_self)

    def submodules(self) -> List[T]:
        """Return a list of submodules."""
        module_subtrees = [
            getattr(self, name)
            for name, kind in self._pax.name_to_kind.items()
            if kind == PaxKind.MODULE
        ]

        is_module = lambda x: isinstance(x, BaseModule)
        submods, _ = jax.tree_flatten(module_subtrees, is_leaf=is_module)
        return [v for v in submods if is_module(v)]

    def _assert_not_shared_module(self):
        """Shared module is not allowed."""
        shared_module = _find_shared_module(self)
        if shared_module is not None:
            raise ValueError(
                f"The module `{shared_module}` is shared between two nodes of the pytree.\n"
                f"This is not allowed to prevent potential silence bugs."
            )

    def _assert_not_shared_weight(self):
        """Shared weight is not allowed."""
        leaves = jax.tree_leaves(self)
        leaf_ids = set()
        for leaf in leaves:
            if id(leaf) in leaf_ids:
                raise ValueError(
                    f"Detected a shared ndarray. This is not allowed.\n"
                    f"Shape={leaf.shape}\n"
                    f"Dtype={leaf.dtype}\n"
                    f"Value={leaf}",
                )
            leaf_ids.add(id(leaf))


def _find_shared_module(module: BaseModule):
    """Find shared module.

    - Return the first module that is shared by two nodes of the pytree.
    - Return `None` if there is no shared module.
    """

    def _get_all_modules(mod: BaseModule, lst: List):
        lst.append(mod)
        for m in mod.submodules():
            _get_all_modules(m, lst)

    mods = []
    _get_all_modules(module, mods)
    module_ids = set()
    for m in mods:
        if id(m) in module_ids:
            return m
        module_ids.add(id(m))

    return None


def _has_module_node(value):
    leaves = jax.tree_flatten(value, is_leaf=lambda x: isinstance(x, BaseModule))[0]
    for leaf in leaves:
        if isinstance(leaf, BaseModule):
            return True

    return False



"""RngSeq module."""

from typing import Optional, Sequence, Union

import jax
import jax.numpy as jnp
import numpy as np

from ..core import StateModule, rng


class RngSeq(StateModule):
    """A module which generates an infinite sequence of rng keys."""

    _rng_key: rng.KeyArray

    def __init__(
        self, seed: Optional[int] = None, rng_key: Optional[rng.KeyArray] = None
    ):
        """Initialize a random key sequence.

        **Note**: ``rng_key`` has a higher priority than ``seed``.

        Arguments:
            seed: an integer seed.
            rng_key: a jax random key.
        """
        super().__init__()
        if rng_key is not None:
            rng_key_ = rng_key
        elif seed is not None:
            rng_key_ = jax.random.PRNGKey(seed)
        else:
            rng_key_ = rng.next_rng_key()

        if isinstance(rng_key_, (np.ndarray, jnp.ndarray)):
            self._rng_key = rng_key_
        else:
            raise ValueError("Impossible")

    def next_rng_key(
        self, num_keys: int = 1
    ) -> Union[rng.KeyArray, Sequence[rng.KeyArray]]:
        """Return the next random key of the sequence.

        **Note**:

        * Return a key if ``num_keys`` is ``1``,
        * Return a list of keys if ``num_keys`` is greater than ``1``.
        * This is not a deterministic sequence if values of ``num_keys`` are mixed randomly.

        Arguments:
            num_keys: return more than one key.
        """
        new_rng_key, *rng_keys = jax.random.split(self._rng_key, num_keys + 1)

        # only update internal state in `train` mode.
        if self.training:
            self._rng_key = new_rng_key
        if num_keys == 1:
            return rng_keys[0]
        else:
            return rng_keys



import math
from typing import Sequence

import jax
import jax.numpy as jnp
import pax
from pax.nets import Transformer


def positional_encoding(x):
    _, L, D = x.shape
    position = jnp.arange(0, L, dtype=x.dtype)[:, None]
    div_term = jnp.exp(jnp.arange(0, D, 2, dtype=x.dtype) * (-math.log(10_000.0) / D))
    x1 = jnp.sin(position * div_term[None, :])
    x2 = jnp.cos(position * div_term[None, :])
    x_pos = jnp.concatenate((x1, x2), axis=-1)
    return x + x_pos[None, :, :]


class LM(pax.Module):
    """A Transformer language model."""

    transformer: Transformer
    embed: pax.Module
    output: pax.Module

    vocab_size: int
    hidden_dim: int

    def __init__(
        self, vocab_size: int, hidden_dim: int, num_layers: int, dropout: float = 0.1
    ):
        """
        Arguments:
            vocab_size: int, size of the alphabet.
            hidden_dim: int, hidden dim.
            num_layers: int, num transformer blocks.
        """
        super().__init__()
        self.vocab_size = vocab_size
        self.hidden_dim = hidden_dim
        self.embed = pax.nn.Embed(
            vocab_size,
            hidden_dim,
            w_init=jax.nn.initializers.variance_scaling(
                1.0, mode="fan_out", distribution="normal"
            ),
        )
        self.transformer = Transformer(
            hidden_dim, hidden_dim // 64, num_layers, dropout_rate=dropout
        )
        self.output = pax.nn.Linear(hidden_dim, vocab_size)

    def __call__(self, x):
        x = self.embed(x)
        x = positional_encoding(x)
        x = self.transformer(x)
        logits = self.output(x)
        return logits

    def inference(self, prompt: Sequence[int] = (), length=1024, train_seq_len=256):
        def step(inputs, _):
            logits = self(inputs)
            x = jnp.argmax(logits[:, -1], axis=-1)
            next_inputs = jnp.concatenate((inputs[:, 1:], x[:, None]), axis=-1)
            return next_inputs, x

        if len(prompt) > train_seq_len:
            inputs = prompt[-train_seq_len:]
        else:
            inputs = prompt
        pad_len = train_seq_len - len(inputs)
        padded_inputs = [0] * pad_len + inputs
        x = jnp.array([padded_inputs], dtype=jnp.int32)
        L = length - len(prompt)
        _, out = pax.scan(step, x, None, length=L, time_major=False)
        return prompt + out[0].tolist()



import jax
import numpy as np
import tensorflow as tf


def tokenize(text):
    t = [0] + [ord(c) for c in text]  # ASCII, 0 is the [START] token
    return t


def detokenize(tokens):
    text = [chr(t) if t != 0 else "[START]" for t in tokens]
    return "".join(text)


def _device_put_sharded(sharded_tree, devices):
    leaves, treedef = jax.tree_flatten(sharded_tree)
    n = leaves[0].shape[0]
    return jax.device_put_sharded(
        [jax.tree_unflatten(treedef, [l[i] for l in leaves]) for i in range(n)], devices
    )


# Source: https://github.com/deepmind/dm-haiku/blob/8fad8c7503c5f56fa9ea9b53f71b7082704e3a3e/examples/imagenet/dataset.py#L163
def double_buffer(ds, num_devices, steps_per_update):
    """Keeps at least two batches on the accelerator.
    The current GPU allocator design reuses previous allocations. For a training
    loop this means batches will (typically) occupy the same region of memory as
    the previous batch. An issue with this is that it means we cannot overlap a
    host->device copy for the next batch until the previous step has finished and
    the previous batch has been freed.
    By double buffering we ensure that there are always two batches on the device.
    This means that a given batch waits on the N-2'th step to finish and free,
    meaning that it can allocate and copy the next batch to the accelerator in
    parallel with the N-1'th step being executed.
    Args:
      ds: Iterable of batches of numpy arrays.
    Yields:
      Batches of sharded device arrays.
    """
    batch = None
    devices = jax.devices()
    for next_batch in ds:
        assert next_batch is not None
        next_batch = np.reshape(
            next_batch, (num_devices, steps_per_update, -1) + next_batch.shape[1:]
        )
        next_batch = _device_put_sharded(next_batch, devices)
        if batch is not None:
            yield batch
        batch = next_batch
    if batch is not None:
        yield batch


def make_data_loader(data, seq_len, batch_size, num_devices, steps_per_update):
    data_token = tokenize(data)
    data_token = [0] * seq_len + data_token

    tfdata = (
        tf.data.Dataset.from_tensors(data_token)
        .repeat()
        .map(
            lambda x: tf.image.random_crop(x, [seq_len + 1]),
            num_parallel_calls=tf.data.AUTOTUNE,
        )
        .batch(batch_size)
        .prefetch(tf.data.AUTOTUNE)
        .as_numpy_iterator()
    )

    return double_buffer(tfdata, num_devices, steps_per_update)



"""GroupNorm module."""

# This file is an adaptation from
# https://raw.githubusercontent.com/deepmind/dm-haiku/main/haiku/_src/group_norm.py
# which is under Apache License, Version 2.0.


import collections
from typing import Callable, Optional, Sequence, Union

import jax
import jax.numpy as jnp

from ..core import ParameterModule
from ..core.rng import KeyArray, next_rng_key

# Copyright 2020 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


class GroupNorm(ParameterModule):
    r"""Group normalization module.

    This applies group normalization to the x. This involves splitting the
    channels into groups before calculating the mean and variance. The default
    behavior is to compute the mean and variance over the spatial dimensions and
    the grouped channels. The mean and variance will never be computed over the
    created groups axis.

    It transforms the input ``x`` into:

    .. math::

       \d{outputs} = \d{scale} \dfrac{x - \mu}{\sigma + \epsilon} + \d{offset}

    Where :math:`\mu` and :math:`\sigma` are respectively the mean and standard
    deviation of ``x``.

    There are many different variations for how users want to manage scale and
    offset if they require them at all. These are:

      - No ``scale``/``offset`` in which case ``create_*`` should be set to
        ``False`` and ``scale``/``offset`` aren't passed when the module is
        called.
      - Trainable ``scale``/``offset`` in which case create_* should be set to
        ``True`` and again ``scale``/``offset`` aren't passed when the module is
        called. In this case this module creates and owns the scale/offset
        parameters.
      - Externally generated ``scale``/``offset``, such as for conditional
        normalization, in which case ``create_*`` should be set to ``False`` and
        then the values fed in at call time.
    """

    scale: jnp.ndarray
    offset: jnp.ndarray

    def __init__(
        self,
        groups: int,
        num_channels: int,
        axis: Union[int, slice, Sequence[int]] = slice(1, None),
        create_scale: bool = True,
        create_offset: bool = True,
        eps: float = 1e-5,
        scale_init: Optional[Callable] = None,
        offset_init: Optional[Callable] = None,
        data_format: str = "channels_last",
        *,
        rng_key: Optional[KeyArray] = None,
        name: Optional[str] = None,
    ):
        """Constructs a ``GroupNorm`` module.

        Args:
            groups: number of groups to divide the channels by. The number of channels
                must be divisible by this.
            num_channels: number of channels.
            axis: ``int``, ``slice`` or sequence of ints representing the axes which
                should be normalized across. By default this is all but the first
                dimension. For time series data use `slice(2, None)` to average over the
                none Batch and Time data.
            create_scale: whether to create a trainable scale per channel applied
                after the normalization.
            create_offset: whether to create a trainable offset per channel applied
                after normalization and scaling.
            eps: Small epsilon to add to the variance to avoid division by zero.
                Defaults to ``1e-5``.
            scale_init: Optional initializer for the scale parameter. Can only be set
                if ``create_scale=True``. By default scale is initialized to ``1``.
            offset_init: Optional initializer for the offset parameter. Can only be
                set if ``create_offset=True``. By default offset is initialized to
                ``0``.
            data_format: The data format of the input. Can be either
                ``channels_first``, ``channels_last``, ``N...C`` or ``NC...``. By
                default it is ``channels_last``.
            name: Name of the module.
        """
        super().__init__(name=name)

        if isinstance(axis, slice):
            self.axis = axis
        elif isinstance(axis, int):
            self.axis = (axis,)
        elif isinstance(axis, collections.abc.Iterable) and all(
            isinstance(ax, int) for ax in axis
        ):
            self.axis = axis
        else:
            raise ValueError("`axis` should be an int, slice or iterable of ints.")

        self.groups = groups
        self.eps = eps
        self.data_format = data_format
        if data_format in ["channels_first", "NC..."]:
            self.channel_index = 1
        elif data_format in ["channels_last", "N...C"]:
            self.channel_index = -1
        else:
            raise ValueError(f"Data format `{data_format}` is not supported.")

        self.create_scale = create_scale
        self.create_offset = create_offset
        self.scale_init = scale_init or jax.nn.initializers.ones
        self.offset_init = offset_init or jax.nn.initializers.zeros

        param_shape = [num_channels]
        rng_key = next_rng_key() if rng_key is None else rng_key
        rng1, rng2 = jax.random.split(rng_key)
        if create_scale:
            self.scale = self.scale_init(rng1, param_shape)
        if create_offset:
            self.offset = self.offset_init(rng2, param_shape)

    def __call__(
        self,
        x: jnp.ndarray,
        scale: Optional[jnp.ndarray] = None,
        offset: Optional[jnp.ndarray] = None,
    ) -> jnp.ndarray:
        """Returns normalized inputs.

        Args:
          x: An n-D tensor of the ``data_format`` specified in the constructor
            on which the transformation is performed.
          scale: A tensor up to n-D. The shape of this tensor must be broadcastable
            to the shape of ``x``. This is the scale applied to the normalized
            x. This cannot be passed in if the module was constructed with
            ``create_scale=True``.
          offset: A tensor up to n-D. The shape of this tensor must be broadcastable
            to the shape of ``x``. This is the offset applied to the normalized
            ``x``. This cannot be passed in if the module was constructed with
            ``create_offset=True``.

        Returns:
          An n-d tensor of the same shape as x that has been normalized.
        """
        if self.create_scale and scale is not None:
            raise ValueError("Cannot pass `scale` at call time if `create_scale=True`.")

        if self.create_offset and offset is not None:
            raise ValueError(
                "Cannot pass `offset` at call time if `create_offset=True`."
            )

        channels = x.shape[self.channel_index]
        if channels % self.groups != 0:
            raise ValueError(
                "The number of channels must be divisible by the number of groups, "
                f"was channels={channels}, groups={self.groups}"
            )

        if self.create_scale:
            scale = self.scale

        if self.create_offset:
            offset = self.offset

        first_input_shape, group_shape, axis = self._compute_shape_and_axis(x, channels)

        x = x.reshape(group_shape)
        mean = jnp.mean(x, axis, keepdims=True)
        # TODO(tycai): Consider faster but less precise variance formulation.
        var = jnp.var(x, axis, keepdims=True)
        x = (x - mean) * jax.lax.rsqrt(var + self.eps)
        x = x.reshape(first_input_shape)

        if scale is not None:
            scale = jax.lax.broadcast_to_rank(scale, x.ndim)
            x = x * scale

        if offset is not None:
            offset = jax.lax.broadcast_to_rank(offset, x.ndim)
            x = x + offset

        return x

    def _compute_shape_and_axis(self, x: jnp.ndarray, channels: int):
        """Make it compatible with haiku code."""
        rank = x.ndim

        # Turns slice into list of axis
        if isinstance(self.axis, slice):
            axes = tuple(range(rank))
            axis = axes[self.axis]
        else:
            axis = self.axis

        if self.channel_index == -1:
            axis = tuple(a if a != rank - 1 else a + 1 for a in axis)
            group_shape = (-1,) + x.shape[1:-1] + (self.groups, channels // self.groups)
        else:
            assert self.channel_index == 1
            axis = tuple(a if a == 0 else a + 1 for a in axis)
            group_shape = (-1, self.groups, channels // self.groups) + x.shape[2:]

        first_input_shape = (-1,) + x.shape[1:]
        return first_input_shape, group_shape, axis



"""Experimental API"""


from pax._src.core import Flattener, LazyModule
from pax._src.utils import (
    apply_scaled_gradients,
    default_mp_policy,
    save_weights_to_dict,
    load_weights_from_dict,
)

from . import graph

__all__ = (
    "apply_scaled_gradients",
    "default_mp_policy",
    "Flattener",
    "graph",
    "LazyModule",
    "load_weights_from_dict",
    "save_weights_to_dict",
)



"""Public nn modules."""

from pax._src.nn import (
    EMA,
    GRU,
    LSTM,
    BatchNorm1D,
    BatchNorm2D,
    Conv1D,
    Conv1DTranspose,
    Conv2D,
    Conv2DTranspose,
    Dropout,
    Embed,
    GroupNorm,
    GRUState,
    Identity,
    Lambda,
    LayerNorm,
    Linear,
    LSTMState,
    MultiHeadAttention,
    RngSeq,
    Sequential,
    avg_pool,
    max_pool,
)

__all__ = (
    "avg_pool",
    "BatchNorm1D",
    "BatchNorm2D",
    "Conv1D",
    "Conv1DTranspose",
    "Conv2D",
    "Conv2DTranspose",
    "Dropout",
    "EMA",
    "Embed",
    "GroupNorm",
    "GRU",
    "GRUState",
    "Identity",
    "Lambda",
    "LayerNorm",
    "Linear",
    "LSTM",
    "LSTMState",
    "max_pool",
    "MultiHeadAttention",
    "RngSeq",
    "Sequential",
)



import os
import random

import librosa
import numpy as np


def data_loader(
    batch_size: int,
    n_mels: int,
    n_fft: int,
    hop_length: int,
    win_length: int,
    sample_rate: int,
    fmin: int,
    fmax: int,
    mu: int,
    n_frames: int,
    split="train",
):
    if not os.path.exists("/tmp/wave_gru_clip.wav"):
        os.system("bash /tmp/prepare_clip.sh")

    wav, _ = librosa.load("/tmp/wave_gru_clip.wav", sr=sample_rate)

    L = len(wav) * 9 // 10
    if split == "train":
        wav = wav[:L]
    else:
        wav = wav[L:]

    mel = librosa.feature.melspectrogram(
        n_mels=n_mels,
        y=wav,
        sr=sample_rate,
        n_fft=n_fft,
        hop_length=hop_length,
        win_length=win_length,
        fmin=fmin,
        fmax=fmax,
        center=False,
    )

    mel = mel.T

    logmel = np.log(1e-3 + mel)
    mu_wav = librosa.mu_compress(wav, mu=mu, quantize=True) + mu // 2

    if split == "test":
        yield (logmel, mu_wav)
        return

    batch = []
    while True:
        left = random.randint(8, logmel.shape[0] - n_frames - 9)
        right = left + n_frames
        cond = logmel[(left - 8) : (right + 1)]  # padding purposes
        x = mu_wav[left * hop_length : right * hop_length + 1]
        batch.append((cond, x))
        if len(batch) == batch_size:
            conds, xs = zip(*batch)
            conds = np.array(conds)
            xs = np.array(xs, dtype=np.int16)
            yield (conds, xs)
            batch = []



"""Public nets."""

from pax._src.nets import (
    ResNet18,
    ResNet34,
    ResNet50,
    ResNet101,
    ResNet152,
    ResNet200,
    Transformer,
)

__all__ = (
    "ResNet18",
    "ResNet34",
    "ResNet50",
    "ResNet101",
    "ResNet152",
    "ResNet200",
    "Transformer",
)



import os

import tensorflow as tf
import tensorflow_datasets as tfds

### load celeb_a dataset

# This is a hack to use a custom link to celeb-a dataset in tensorflow-datasets.
# replace the ``tfds.image.CelebA._split_generators`` method by the following method
# which uses our custom links.

IMG_ALIGNED_DATA = (
    "https://drive.google.com/uc?export=download&"
    "id=1iQRFaGXRiPBd-flIm0u-u8Jy6CfJ_q6j"
)

EVAL_LIST = (
    "https://drive.google.com/uc?export=download&"
    "id=1ab9MDLOblszbKKXoDe8jumFsSkn6lIX1"
)
# Landmark coordinates: left_eye, right_eye etc.
LANDMARKS_DATA = (
    "https://drive.google.com/uc?export=download&"
    "id=1y8qfK-jaq1QWl9v_n_mBNIMu5-h3UXK4"
)

# Attributes in the image (Eyeglasses, Mustache etc).
ATTR_DATA = (
    "https://drive.google.com/uc?export=download&"
    "id=1BPfcVuIqrAsJAgG40-XGWU7g2wmmQU30"
)


def _split_generators(self, dl_manager):
    downloaded_dirs = dl_manager.download(
        {
            "img_align_celeba": IMG_ALIGNED_DATA,
            "list_eval_partition": EVAL_LIST,
            "list_attr_celeba": ATTR_DATA,
            "landmarks_celeba": LANDMARKS_DATA,
        }
    )

    # Load all images in memory (~1 GiB)
    # Use split to convert: `img_align_celeba/000005.jpg` -> `000005.jpg`
    all_images = {
        os.path.split(k)[-1]: img
        for k, img in dl_manager.iter_archive(downloaded_dirs["img_align_celeba"])
    }

    return [
        tfds.core.SplitGenerator(
            name=tfds.Split.TRAIN,
            gen_kwargs={
                "file_id": 0,
                "downloaded_dirs": downloaded_dirs,
                "downloaded_images": all_images,
            },
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.VALIDATION,
            gen_kwargs={
                "file_id": 1,
                "downloaded_dirs": downloaded_dirs,
                "downloaded_images": all_images,
            },
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.TEST,
            gen_kwargs={
                "file_id": 2,
                "downloaded_dirs": downloaded_dirs,
                "downloaded_images": all_images,
            },
        ),
    ]


img_mean = 0.5
img_scale = 0.5
image_size = 64  # size of input image: 64x64

tfds.image.CelebA._split_generators = _split_generators


def load_celeb_a():
    ds = tfds.load("celeb_a")

    def img_ops(x):
        img = tf.cast(x["image"], tf.float32) / 255.0
        img = tf.image.resize(
            img, (image_size * 2, image_size), preserve_aspect_ratio=True
        )
        img = tf.image.crop_to_bounding_box(img, 7, 0, 64, 64)
        img = (img - img_mean) / img_scale
        return img

    dataset = (
        ds["train"].concatenate(ds["validation"]).concatenate(ds["test"]).map(img_ops)
    )
    return dataset



"""PAX Module"""

from .graph_module import GraphModule, InputNode, build_graph_module
from .mixed_precision import apply_mp_policy, unwrap_mp_policy
from .module import Module, PaxKind
from .module_and_value import module_and_value
from .pure import pure
from .transforms import (
    enable_eval_mode,
    enable_train_mode,
    freeze_parameters,
    select_parameters,
    select_states,
    unfreeze_parameters,
    update_parameters,
    update_states,
)
from .utility_modules import Flattener, LazyModule, ParameterModule, StateModule
from .utils import assert_structure_equal



"""Sequential module."""

from typing import Optional, Tuple, TypeVar

from ..core import Module
from .lambda_module import Lambda

T = TypeVar("T", bound=Module)


class Sequential(Module):
    """Execute layers in order.

    Support pax.Module (callable pytree) and any jax functions.

    For example:

    >>> net = pax.nn.Sequential(
    ...              pax.nn.Linear(2, 32),
    ...              jax.nn.relu,
    ...              pax.nn.Linear(32, 3)
    ... )
    >>> print(net.summary())
    Sequential
    ├── Linear(in_dim=2, out_dim=32, with_bias=True)
    ├── x => relu(x)
    └── Linear(in_dim=32, out_dim=3, with_bias=True)
    >>> x = jnp.empty((3, 2))
    >>> y = net(x)
    >>> y.shape
    (3, 3)
    """

    # Note: we cannot mix pax.Module and jax functions (e.g., jax.nn.relu) in the same list.
    # therefore, we have to convert a jax function to ``Lambda`` module first.
    modules: Tuple[Module, ...]

    def __init__(self, *layers, name: Optional[str] = None):
        """Create a Sequential module."""
        super().__init__(name=name)
        self.modules = tuple(
            (f if isinstance(f, Module) else Lambda(f)) for f in layers
        )

    def __call__(self, x):
        """Call layers in order."""
        for f in self.modules:
            x = f(x)
        return x

    def __getitem__(self, index: int) -> T:
        """Get an item from the `modules` list."""
        return self.modules[index]

    def set(self: T, index: int, value) -> T:
        """Set an item to the `modules` list."""
        if not isinstance(value, Module):
            value = Lambda(value)

        modules = list(self.modules)
        modules[index] = value
        return super().replace(modules=tuple(modules))

    def __rshift__(self, other: Module):
        return Sequential(*self.modules, other)



"""Embed module."""

from typing import Callable, Optional

import jax
import jax.numpy as jnp

from ..core import ParameterModule
from ..core.rng import KeyArray, next_rng_key


class Embed(ParameterModule):
    """Embed module maps integer values to real vectors.
    The embedded vectors are trainable.
    """

    weight: jnp.ndarray
    vocab_size: int
    embed_dim: int

    def __init__(
        self,
        vocab_size: int,
        embed_dim: int,
        w_init: Optional[Callable] = None,
        *,
        rng_key: Optional[KeyArray] = None,
        name: Optional[str] = None
    ):
        """
        An embed module.

        Arguments:
            vocab_size: the number of embedded vectors.
            embed_dim: the size of embedded vectors.
            w_init: weight initializer. Default: `truncated_normal`.
            name: module name.
        """

        super().__init__(name=name)

        self.vocab_size = vocab_size
        self.embed_dim = embed_dim
        shape = [vocab_size, embed_dim]

        if w_init is None:
            w_init = jax.nn.initializers.normal()

        if rng_key is None:
            rng_key = next_rng_key()

        self.weight = w_init(rng_key, shape)

    def __call__(self, x: jnp.ndarray):
        """Return embedded vectors indexed by ``x``."""
        return self.weight[(x,)]

    def __repr__(self):
        info = {"vocab_size": self.vocab_size, "embed_dim": self.embed_dim}
        return self._repr(info)



"""Experimental graph API"""


from pax._src.core.graph_module import GraphModule, InputNode, Node, build_graph_module

__all__ = (
    "build_graph_module",
    "GraphModule",
    "InputNode",
    "Node",
)



"""Recurrent modules."""

from typing import Callable, NamedTuple, Optional, Tuple

import jax
import jax.numpy as jnp

from ..core import Module
from ..core.rng import KeyArray, next_rng_key
from .linear import Linear


class LSTMState(NamedTuple):
    """LSTMState."""

    hidden: jnp.ndarray
    cell: jnp.ndarray


class GRUState(NamedTuple):
    """GRUState."""

    hidden: jnp.ndarray


class RNN(Module):
    """Base class for all recurrent modules."""

    def __init__(self, name: Optional[str] = None):
        super().__init__(name=name)

    def initial_state(self, batch_size):
        raise NotImplementedError()


class LSTM(RNN):
    """Long Short Term Memory (LSTM) RNN module."""

    input_dim: int
    hidden_dim: int

    weight: jnp.ndarray
    bias: jnp.ndarray

    def __init__(
        self,
        input_dim: int,
        hidden_dim: int,
        w_init: Optional[Callable] = None,
        forget_gate_bias: float = 0.0,
        *,
        rng_key: KeyArray = None,
        name: Optional[str] = None
    ):
        """Create a LSTM module.

        Arguments:
            input_dim: The input dimension.
            hidden_dim: The number of LSTM cells.
            w_init: weight initializer.
            forget_gate_bias: Prefer forget. Default `0`.
            rng_key: random key.
            name: module name.
        """

        super().__init__(name=name)
        self.input_dim = input_dim
        self.hidden_dim = hidden_dim
        self.forget_gate_bias = forget_gate_bias

        self.fc = Linear(
            (input_dim + hidden_dim),
            4 * hidden_dim,
            rng_key=rng_key,
            name="lstm_fc",
            w_init=w_init,
        )

    def __call__(
        self,
        state: LSTMState,
        x: jnp.ndarray,
    ) -> Tuple[LSTMState, jnp.ndarray]:
        """Do a single lstm step.


        Arguments:
            state: The current LSTM state.
            x: The input.
        """
        xh = jnp.concatenate((x, state.hidden), axis=-1)
        gated = self.fc(xh)
        i, g, f, o = jnp.split(gated, 4, axis=-1)
        f = jax.nn.sigmoid(f + self.forget_gate_bias)
        c = f * state.cell + jax.nn.sigmoid(i) * jnp.tanh(g)
        h = jax.nn.sigmoid(o) * jnp.tanh(c)
        return LSTMState(h, c), h

    def __repr__(self):
        info = {"input_dim": self.input_dim, "hidden_dim": self.hidden_dim}
        return self._repr(info)

    def initial_state(self, batch_size) -> LSTMState:
        shape = (batch_size, self.hidden_dim)
        hidden = jnp.zeros(shape=shape, dtype=jnp.float32)
        cell = jnp.zeros(shape=shape, dtype=jnp.float32)
        return LSTMState(hidden=hidden, cell=cell)


class GRU(RNN):
    """This class implements the "fully gated unit" GRU.

    Reference: https://en.wikipedia.org/wiki/Gated_recurrent_unit
    """

    input_dim: int
    hidden_dim: int

    def __init__(
        self,
        input_dim: int,
        hidden_dim: int,
        *,
        rng_key: Optional[KeyArray] = None,
        name: Optional[str] = None
    ):
        """Create a GRU module.

        Arguments:
            input_dim: the input size.
            hidden_dim: the number of GRU cells.
        """
        super().__init__(name=name)

        self.input_dim = input_dim
        self.hidden_dim = hidden_dim

        if rng_key is None:
            rng_key = next_rng_key()
        rng_key_1, rng_key_2 = jax.random.split(rng_key, 2)
        self.xh_zr_fc = Linear(
            (input_dim + hidden_dim), hidden_dim * 2, name="xh_to_zr", rng_key=rng_key_1
        )

        self.xh_h_fc = Linear(
            (input_dim + hidden_dim), hidden_dim, name="xh_to_h", rng_key=rng_key_2
        )

    def initial_state(self, batch_size: int) -> GRUState:
        """Create an all zeros initial state."""
        return GRUState(jnp.zeros((batch_size, self.hidden_dim), dtype=jnp.float32))

    def __call__(self, state: GRUState, x) -> Tuple[GRUState, jnp.ndarray]:
        """Do a single gru step.

        Arguments:
            state: The current GRU state.
            x: The input.
        """
        hidden = state.hidden
        xh = jnp.concatenate((x, hidden), axis=-1)
        zr = jax.nn.sigmoid(self.xh_zr_fc(xh))
        z, r = jnp.split(zr, 2, axis=-1)

        xrh = jnp.concatenate((x, r * hidden), axis=-1)
        h_hat = jnp.tanh(self.xh_h_fc(xrh))
        h = (1 - z) * hidden + z * h_hat
        return GRUState(h), h

    def __repr__(self):
        info = {"input_dim": self.input_dim, "hidden_dim": self.hidden_dim}
        return self._repr(info)



"""Linear module."""

from typing import Optional

import jax
import jax.numpy as jnp
import numpy as np

from ..core import ParameterModule
from ..core.rng import KeyArray, next_rng_key


class Linear(ParameterModule):
    """A linear transformation is applied over the last dimension of the input."""

    weight: jnp.ndarray
    bias: jnp.ndarray

    in_dim: int
    out_dim: int
    with_bias: bool

    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        with_bias: bool = True,
        w_init=None,
        b_init=None,
        *,
        name: Optional[str] = None,
        rng_key: KeyArray = None,
    ):
        """
        Arguments:
            in_dim: the number of input features.
            out_dim: the number of output features.
            with_bias: whether to add a bias to the output (default: True).
            w_init: initializer function for the weight matrix.
            b_init: initializer function for the bias.
            rng_key: the key to generate initial parameters.
        """
        super().__init__(name=name)
        self.in_dim = in_dim
        self.out_dim = out_dim
        self.with_bias = with_bias

        rng_key = next_rng_key() if rng_key is None else rng_key
        if w_init is None:
            w_init = jax.nn.initializers.normal(stddev=1.0 / np.sqrt(self.in_dim))
        b_init = jax.nn.initializers.zeros
        rng_key_w, rng_key_b = jax.random.split(rng_key)
        self.weight = w_init(rng_key_w, (in_dim, out_dim))
        if self.with_bias:
            self.bias = b_init(rng_key_b, (out_dim,))

    def __call__(self, x: np.ndarray) -> jnp.ndarray:
        """Applies a linear transformation to the inputs along the last dimension.

        Arguments:
            x: The nd-array to be transformed.
        """
        assert len(x.shape) >= 2, "expecting an input of shape `N...C`"
        x = jnp.dot(x, self.weight)
        if self.with_bias:
            x = x + self.bias
        return x

    def __repr__(self):
        info = {
            "in_dim": self.in_dim,
            "out_dim": self.out_dim,
            "with_bias": self.with_bias,
        }
        return self._repr(info)



from functools import partial

import jax
import jax.numpy as jnp
import librosa
import opax
import pax
import soundfile
from tqdm.auto import tqdm

from data_loader import data_loader
from model import WaveGRU


def loss_fn(model: WaveGRU, inputs):
    logmel, wav = inputs
    input_wav = wav[:, :-1]
    target_wav = wav[:, 1:]
    model, logits = pax.module_and_value(model, check_leaks=False)((logmel, input_wav))
    log_pr = jax.nn.log_softmax(logits, axis=-1)
    target_wave = jax.nn.one_hot(target_wav, num_classes=logits.shape[-1])
    log_pr = jnp.sum(log_pr * target_wave, axis=-1)
    loss = -jnp.mean(log_pr)
    return loss, (loss, model)


def generate_test_sample(step, test_logmel, wave_gru, length, sample_rate, mu):
    generated_mu = wave_gru.eval().inference(test_logmel[None, :length, :])
    generated_mu = jax.device_get(generated_mu)
    synthesized_clip = librosa.mu_expand(
        generated_mu[0] - mu // 2, mu=mu, quantize=True
    )
    file_name = f"/tmp/wave_gru_sample_{step:05d}.wav"
    soundfile.write(
        file_name,
        synthesized_clip,
        samplerate=sample_rate,
    )
    return file_name


def train(
    hidden_dim: int = 512,
    num_training_steps: int = 5_000,
    batch_size: int = 128,
    learning_rate: float = 5e-4,
    sample_rate: int = 16_000,
    max_global_norm: float = 1.0,
    n_fft=1024,
    hop_length=256,
    win_length=1024,
    n_mels=80,
    fmin=0,
    fmax=8000,
    seq_len=2 ** 10,
    n_mu_bits=8,
    log_freq: int = 1000,
    random_seed=42,
):
    pax.seed_rng_key(random_seed)
    mu = 2 ** n_mu_bits - 1
    n_frames = seq_len // hop_length
    wave_gru = WaveGRU(n_mels, hidden_dim)
    print(wave_gru.summary())

    optimizer = opax.chain(
        opax.clip_by_global_norm(max_global_norm),
        opax.adam(learning_rate),
    )(wave_gru.parameters())

    split_loader = partial(
        data_loader,
        batch_size=batch_size,
        n_mels=n_mels,
        n_fft=n_fft,
        hop_length=hop_length,
        win_length=win_length,
        sample_rate=sample_rate,
        mu=mu,
        n_frames=n_frames,
        fmin=fmin,
        fmax=fmax,
    )
    data_iter = split_loader(split="train")
    test_iter = split_loader(split="test")
    test_logmel, _ = next(test_iter)

    update_fn = jax.jit(pax.utils.build_update_fn(loss_fn))
    total_loss = 0.0
    tr = tqdm(range(1, 1 + num_training_steps))
    for step in tr:
        batch = next(data_iter)
        wave_gru, optimizer, loss = update_fn(wave_gru, optimizer, batch)
        total_loss = total_loss + loss

        if step % log_freq == 0:
            loss = total_loss / log_freq
            total_loss = 0.0
            file_name = generate_test_sample(
                step, test_logmel, wave_gru, 1000, sample_rate, mu
            )
            tr.write(
                f"[step {step}]  train loss {loss:.3f}  synthesized clip {file_name}"
            )


if __name__ == "__main__":
    import fire

    fire.Fire(train)



"""BatchNorm modules."""

from typing import Optional, Sequence

import jax
import jax.numpy as jnp

from ..core import Module
from .ema import EMA


class BatchNorm(Module):
    """A Generic BatchNorm Module.

    Normalize a mini-batch of data by subtracting its mean and dividing by its standard deviation.

    Use EMA modules to track the averaged mean and averaged variance for later uses in `eval` mode.
    """

    scale: Optional[jnp.ndarray]
    offset: Optional[jnp.ndarray]

    ema_mean: EMA
    ema_var: EMA

    reduced_axes: Sequence[int]
    create_offset: bool
    create_scale: bool
    eps: float
    data_format: Optional[str]

    def __init__(
        self,
        num_channels: int,
        create_scale: bool = True,
        create_offset: bool = True,
        decay_rate: float = 0.9,
        eps: float = 1e-5,
        data_format: Optional[str] = None,
        reduced_axes=None,
        param_shape=None,
        *,
        name: Optional[str] = None,
    ):
        """Create a new BatchNorm module.

        Arguments:
            num_channels: the number of filters.
            create_scale: create a trainable `scale` parameter.
            create_offset: create a trainable `offset` parameter.
            decay_rate: the decay rate for tracking the averaged mean and the averaged variance.
            eps: a small positive number to avoid divided by zero.
            data_format:  the data format ["NHWC", NCHW", "NWC", "NCW"].
            reduced_axes: list of axes that will be reduced in the `jnp.mean` computation.
            param_shape: the shape of parameters.
        """
        super().__init__(name=name)
        assert 0 <= decay_rate <= 1

        self.num_channels = num_channels
        self.data_format = data_format
        self.create_scale = create_scale
        self.create_offset = create_offset
        self.eps = eps
        self.decay_rate = decay_rate

        self.reduced_axes = tuple(reduced_axes)

        with self.add_parameters():
            if create_scale:
                self.scale = jnp.ones(param_shape, dtype=jnp.float32)
            else:
                self.scale = None
            if create_offset:
                self.offset = jnp.zeros(param_shape, dtype=jnp.float32)
            else:
                self.offset = None

        self.ema_mean = EMA(jnp.zeros_like(self.offset), decay_rate, debias=True)
        self.ema_var = EMA(jnp.zeros_like(self.offset), decay_rate, debias=True)

    def __call__(self, x):
        if self.training:
            batch_mean = jnp.mean(x, axis=self.reduced_axes, keepdims=True)
            batch_mean_of_squares = jnp.mean(
                jnp.square(x), axis=self.reduced_axes, keepdims=True
            )
            batch_var = batch_mean_of_squares - jnp.square(batch_mean)
            self.ema_mean(batch_mean)
            self.ema_var(batch_var)
        else:
            batch_mean = self.ema_mean.averages
            batch_var = self.ema_var.averages

        if self.create_scale:
            scale = self.scale
        else:
            scale = 1.0

        if self.create_offset:
            offset = self.offset
        else:
            offset = 0.0

        inv = scale * jax.lax.rsqrt(batch_var + self.eps)
        x = (x - batch_mean) * inv + offset
        return x

    def __repr__(self):
        info = {
            "num_channels": self.num_channels,
            "create_scale": self.create_scale,
            "create_offset": self.create_offset,
            "data_format": self.data_format,
            "decay_rate": self.decay_rate,
        }
        return self._repr(info)

    def summary(self, return_list: bool = False):
        lines = super().summary(return_list=True)
        if return_list:
            return lines[:1]
        else:
            return lines[0]


class BatchNorm1D(BatchNorm):
    """The 1D version of BatchNorm."""

    def __init__(
        self,
        num_channels: int,
        create_scale: bool = True,
        create_offset: bool = True,
        decay_rate: float = 0.9,
        eps: float = 1e-5,
        data_format: str = "NWC",
        *,
        name: Optional[str] = None,
    ):
        assert data_format in ["NWC", "NCW"], "expecting a correct `data_format`"

        param_shape = [1, 1, 1]
        if data_format == "NWC":
            axis = -1
            reduced_axes = [0, 1]
        else:
            axis = 1
            reduced_axes = [0, 2]
        param_shape[axis] = num_channels

        super().__init__(
            num_channels=num_channels,
            create_scale=create_scale,
            create_offset=create_offset,
            decay_rate=decay_rate,
            eps=eps,
            data_format=data_format,
            param_shape=param_shape,
            reduced_axes=reduced_axes,
            name=name,
        )


class BatchNorm2D(BatchNorm):
    """The 2D version of BatchNorm."""

    def __init__(
        self,
        num_channels: int,
        create_scale: bool = True,
        create_offset: bool = True,
        decay_rate: float = 0.9,
        eps: float = 1e-5,
        data_format: str = "NHWC",
        *,
        name: Optional[str] = None,
    ):
        assert data_format in ["NHWC", "NCHW"], "expecting a correct `data_format`"

        param_shape = [1, 1, 1, 1]
        if data_format == "NHWC":
            axis = -1
            reduced_axes = [0, 1, 2]
        else:
            axis = 1
            reduced_axes = [0, 2, 3]
        param_shape[axis] = num_channels

        super().__init__(
            num_channels=num_channels,
            create_scale=create_scale,
            create_offset=create_offset,
            decay_rate=decay_rate,
            eps=eps,
            data_format=data_format,
            param_shape=param_shape,
            reduced_axes=reduced_axes,
            name=name,
        )



"""Build a module from a directed graph"""

from dataclasses import dataclass
from functools import lru_cache, wraps
from typing import Callable, List, Optional, Tuple, Union

import jax.numpy as jnp

from .module import Module
from .pure import pure
from .utility_modules import Lambda


def identity(x):
    return x


@dataclass(repr=False)
class Node:
    """A node is an object that stores:

    - parents nodes,
    - a PAX module (or a function),
    - and a value.

    For example:

    >>> x = pax.experimental.graph.Node((), lambda x: x, jnp.array(0))
    >>> x.parents, x.fx, x.value
    ((), Lambda..., DeviceArray(0, dtype=int32, weak_type=True))
    """

    parents: Tuple["Node", ...]
    fx: Union[Module, Callable]
    value: jnp.ndarray

    def __post_init__(self):
        if not isinstance(self.fx, Module):
            self.fx = Lambda(self.fx)

    def __rshift__(self, fn):
        """Create a new node by applying `fn` to the node's value.

        Example:

        >>> import jax, pax, jax.numpy as jnp
        >>> from functools import partial
        >>> x = pax.experimental.graph.Node((), lambda x: x, jnp.array(1.))
        >>> y = x >> partial(jax.lax.add, 1.)
        >>> y.value
        DeviceArray(2., dtype=float32, weak_type=True)
        """
        return Node((self,), fn, pure(fn)(self.value))

    def __and__(self: "Node", other: "Node"):
        """Concatenate two nodes to create a tuple.

        Example:

        >>> x = pax.experimental.graph.InputNode(1)
        >>> y = pax.experimental.graph.InputNode(2)
        >>> z = x & y
        >>> z.value
        (1, 2)
        """
        if isinstance(other, CatNode):
            return CatNode((self, *other.parents), identity, (self.value, *other.value))
        else:
            return CatNode((self, other), identity, (self.value, other.value))

    def __or__(self: "Node", other: "Node"):
        """Concatenate two nodes to create a multi args node.

        Example:

        >>> x = pax.experimental.graph.InputNode(1)
        >>> y = pax.experimental.graph.InputNode(2)
        >>> z = (x | y) >> jax.lax.add
        >>> z.value
        DeviceArray(3, dtype=int32, weak_type=True)
        """
        if isinstance(other, ArgsNode):
            return ArgsNode(
                (self, *other.parents), identity, (self.value, *other.value)
            )
        else:
            return ArgsNode((self, other), identity, (self.value, other.value))

    def binary_ops(self, fn, other):
        """Create a new node using a binary operator.

        Example:

        >>> x = pax.experimental.graph.InputNode(1)
        >>> y = pax.experimental.graph.InputNode(2)
        >>> z = x.binary_ops(jax.lax.sub, y)
        >>> z.value
        DeviceArray(-1, dtype=int32, weak_type=True)
        """

        @wraps(fn)
        def bin_ops_fn(xs):
            return fn(xs[0], xs[1])

        return Node((self, other), bin_ops_fn, fn(self.value, other.value))

    @property
    def shape(self):
        """Return the shape of value.

        Example:

        >>> x = pax.experimental.graph.InputNode(jnp.empty((3, 4)))
        >>> x.shape
        (3, 4)
        """
        if hasattr(self, "value") and self.value is not None:
            return self.value.shape
        else:
            return None

    @property
    def dtype(self):
        """Return dtype of value.

        Example:

        >>> x = pax.experimental.graph.InputNode(jnp.empty((1,), dtype=jnp.int32))
        >>> x.dtype
        dtype('int32')
        """
        if hasattr(self, "value") and self.value is not None:
            return self.value.dtype
        else:
            return None

    def __eq__(self, o: object) -> bool:
        return self is o

    def __hash__(self) -> int:
        return id(self)

    def __deepcopy__(self, _):
        raise TypeError("DO NOT COPY")

    def __copy__(self):
        raise TypeError("DO NOT COPY")


class InputNode(Node):
    """An InputNode object represents an input argument of a GraphModule."""

    parents: Tuple[Node, ...]
    fx: Union[Module, Callable]
    value: jnp.ndarray

    def __init__(self, value: jnp.ndarray, fx=lambda x: x):
        """Creata an InputNode object from a value."""
        super().__init__((), fx, value)


class CatNode(Node):
    """Concatenate two nodes to create a new "tuple" node."""

    def __rshift__(self, fn):
        return Node(self.parents, fn, fn(self.value))

    def __and__(self, other: Node):
        if isinstance(other, CatNode):
            return CatNode(
                (*self.parents, *other.parents),
                identity,
                (*self.value, *other.value),
            )
        else:
            return CatNode((*self.parents, other), identity, (*self.value, other.value))


class ArgsNode(Node):
    """Concatenate two nodes to create a multi output node."""

    def __rshift__(self, fn):
        """Create a new node with multiple arguments."""

        @wraps(fn)
        def multi_args_fn(xs):
            return fn(*xs)

        assert isinstance(self.value, tuple)
        return Node((self,), multi_args_fn, multi_args_fn(self.value))

    def __or__(self, other: Node):
        if isinstance(other, ArgsNode):
            return ArgsNode(
                (*self.parents, *other.parents),
                identity,
                (*self.value, *other.value),
            )
        else:
            return ArgsNode(
                (*self.parents, other), identity, (*self.value, other.value)
            )


class GraphModule(Module):
    """A module that uses a directed graph to represent its computation."""

    output_node: Node
    modules: List[Module]

    def __init__(self, inputs, output, name: Optional[str] = None):
        super().__init__(name=name)

        self.register_module("modules", [])

        @lru_cache(maxsize=None)
        def transform(node: Node):
            p = tuple(transform(parent) for parent in node.parents)
            if node in inputs:
                idx = inputs.index(node)
                return InputNode(value=None, fx=lambda xs: xs[idx])
            elif isinstance(node.fx, Module):
                mod_idx = len(self.modules)
                self.modules.append(node.fx)
                fx = lambda mods, xs: mods[mod_idx](xs)
                fx.__name__ = f"F[{mod_idx}]"
                return Node(p, fx, None)
            else:
                raise RuntimeError("Impossible")

        self.output_node = transform(output)

    def __call__(self, *xs):
        @lru_cache(maxsize=None)
        def run(node: Node):
            if isinstance(node, InputNode):
                return node.fx(xs)
            else:
                ii = tuple(run(p) for p in node.parents)
                if len(node.parents) == 1:
                    (ii,) = ii
                return node.fx(self.modules, ii)

        return run(self.output_node)


def build_graph_module(func):
    """Build a graph module from a forward function.

    Example:

    >>> def residual_forward(x):
    ...     y = x >> pax.nn.Linear(x.shape[-1], x.shape[-1])
    ...     y >>= jax.nn.relu
    ...     z = (x | y) >> jax.lax.add
    ...     return z
    ...
    >>> from pax.experimental.graph import build_graph_module
    >>> net = build_graph_module(residual_forward)(jnp.empty((3, 8)))
    >>> print(net.summary())
    GraphModule
    ├── Linear(in_dim=8, out_dim=8, with_bias=True)
    ├── x => relu(x)
    ├── x => identity(x)
    └── x => add(x)
    """

    def builder_fn(*inputs):
        inputs = tuple(InputNode(x) for x in inputs)
        output = func(*inputs)
        assert isinstance(output, Node)
        mod = GraphModule(inputs, output)
        del inputs, output
        return mod

    return builder_fn



"""PAX module."""

from typing import Any, Dict, List, Optional, Tuple, TypeVar, Union

import jax
import jax.numpy as jnp
import jax.tree_util

from .base import BaseModule, PaxKind
from .module_and_value import module_and_value
from .threading_local import allow_mutation
from .transforms import (
    enable_eval_mode,
    enable_train_mode,
    select_parameters,
    update_parameters,
    update_pytree,
)

T = TypeVar("T", bound="Module")
M = TypeVar("M")
TreeDef = Any


class Module(BaseModule):
    """The base class for all PAX modules.

    Example:

    >>> class Counter(pax.Module):
    ...     def __init__(self):
    ...         super().__init__()
    ...         with self.add_states():
    ...             self.count = jnp.array(0)
    ...
    ...     def step(self, x):
    ...         self.count += 1
    """

    name: Optional[str] = None

    def __init__(self, name: Optional[str] = None):
        """Initialize module.

        Set module name.

        >>> linear = pax.nn.Linear(3, 3, name="input_layer")
        >>> print(linear)
        (input_layer) Linear(in_dim=3, out_dim=3, with_bias=True)
        """
        super().__init__()
        self.name = name

    @property
    def training(self) -> bool:
        """Return `True` if a module is in training mode.

        >>> net = pax.nn.Linear(1, 1)
        >>> net.training
        True
        >>> net = net.eval()
        >>> net.training
        False
        """
        return self._pax.training

    def register_parameter(self, name: str, value: Any):
        """Register a trainable parameter.

        >>> class M(pax.Module):
        ...     def __init__(self):
        ...         super().__init__()
        ...         self.register_parameter("weight", jnp.array(0.))
        ...
        >>> m = M()
        >>> m._pax
        PaxModuleInfo[training=True, nodes={weight:PARAMETER}]
        >>> m.weight
        DeviceArray(0., dtype=float32, weak_type=True)
        """
        self.register_subtree(name, value, PaxKind.PARAMETER)

    def register_state(self, name: str, value: Any):
        """Register a non-trainable state.

        >>> class Counter(pax.Module):
        ...     def __init__(self):
        ...         super().__init__()
        ...         self.register_state("count", jnp.array(0))
        ...
        >>> counter = Counter()
        >>> counter._pax
        PaxModuleInfo[training=True, nodes={count:STATE}]
        >>> counter.count
        DeviceArray(0, dtype=int32, weak_type=True)
        """
        self.register_subtree(name, value, PaxKind.STATE)

    def register_modules(self, name: str, value: Any):
        """Register a module subtree."""
        self.register_subtree(name, value, PaxKind.MODULE)

    register_parameters = register_parameter
    register_states = register_state
    register_module = register_modules

    def copy(self: T) -> T:
        """Return a copy of the current module."""
        leaves, treedef = jax.tree_flatten(self)
        return jax.tree_unflatten(treedef, leaves)

    def train(self: T) -> T:
        """Return a module in training mode."""
        return enable_train_mode(self)

    def eval(self: T) -> T:
        """Return a module in evaluation mode."""
        return enable_eval_mode(self)

    def parameters(self: T) -> T:
        """Return trainable parameters."""
        return select_parameters(self)

    def update_parameters(self: T, params: T) -> T:
        """Return a new module with updated parameters."""
        return update_parameters(self, params=params)

    def replace(self: T, **kwargs) -> T:
        """Return a new module with some attributes replaced.

        >>> net = pax.nn.Linear(2, 2)
        >>> net = net.replace(bias=jnp.zeros((2,)))
        """

        mod = self.copy()
        with allow_mutation(mod):
            for name, value in kwargs.items():
                assert hasattr(mod, name)
                setattr(mod, name, value)
            # pylint: disable=protected-access
            mod._find_and_register_pytree(PaxKind.MODULE)

        mod.scan_bugs()
        return mod

    # inspired by patrick-kidger/equinox `tree_at`
    def replace_node(self: T, node: jnp.ndarray, value: jnp.ndarray) -> T:
        """Replace a node of the pytree by a new value.

        Example:

        >>> mod = pax.nn.Sequential(
        ...     pax.nn.Linear(2,2),
        ...     jax.nn.relu
        ... )
        >>> mod = mod.replace_node(mod[0].weight, jnp.zeros((2, 3)))
        >>> print(mod[0].weight.shape)
        (2, 3)
        """
        leaves, tree_def = jax.tree_flatten(self, is_leaf=lambda x: x is node)
        count = sum(1 if x is node else 0 for x in leaves)

        if count != 1:
            raise ValueError(f"The node `{node}` appears {count} times in the module.")

        # replace `node` by value
        new_leaves = [value if v is node else v for v in leaves]
        mod: T = jax.tree_unflatten(tree_def, new_leaves)
        mod.scan_bugs()
        return mod

    def scan_bugs(self: T) -> T:
        """Scan the module for potential bugs."""

        # scan for shared module/weight.
        self._assert_not_shared_module()
        self._assert_not_shared_weight()

        def _scan_field_fn(mod: T) -> T:
            assert isinstance(mod, Module)
            # pylint: disable=protected-access
            mod._scan_fields(mod._class_fields())
            # pylint: disable=protected-access
            mod._scan_fields(mod.__dict__.keys())
            return mod

        self.apply(_scan_field_fn)
        return self

    def __mod__(self: T, args: Union[Any, Tuple]) -> Tuple[T, Any]:
        """An alternative to `pax.module_and_value`.

        >>> bn = pax.nn.BatchNorm1D(3)
        >>> x = jnp.ones((5, 8, 3))
        >>> bn, y = bn % x
        >>> bn
        BatchNorm1D(num_channels=3, ...)
        """
        assert callable(self)

        if isinstance(args, tuple):
            return module_and_value(self)(*args)
        else:
            return module_and_value(self)(args)

    def __or__(self: T, other: T) -> T:
        """Merge two modules.

        >>> a = pax.nn.Linear(2, 2)
        >>> b = pax.nn.Linear(2, 2)
        >>> c = a | b
        >>> d = b | a
        >>> c == d
        False
        """
        return update_pytree(self, other=other)

    def __invert__(self: T) -> T:
        return self.parameters()

    def map(self: T, func, *mods) -> T:
        return jax.tree_map(func, self, *mods)

    def _repr(self, info: Optional[Dict[str, Any]] = None) -> str:
        name = f"({self.name}) " if self.name is not None else ""
        cls_name = self.__class__.__qualname__
        if info is None:
            return f"{name}{cls_name}"
        else:
            lst_info = [f"{k}={v}" for (k, v) in info.items() if v is not None]
            str_info = ", ".join(lst_info)
            return f"{name}{cls_name}({str_info})"

    def __repr__(self) -> str:
        return self._repr()

    def summary(self, return_list: bool = False) -> Union[str, List[str]]:
        """Summarize a module as a tree of its submodules.

        Arguments:
            return_list: return a list of lines instead of a joined string.

        >>> net = pax.nn.Sequential(pax.nn.Linear(2, 3), jax.nn.relu, pax.nn.Linear(3, 4))
        >>> print(net.summary())
        Sequential
        ├── Linear(in_dim=2, out_dim=3, with_bias=True)
        ├── x => relu(x)
        └── Linear(in_dim=3, out_dim=4, with_bias=True)
        """

        output = [self.__repr__()]
        if output[0] is None:
            raise ValueError(
                f"The `{self.__class__}.__repr__` method returns a `None` value."
            )
        submodules: List[Module] = self.submodules()

        def indent(lines: List[str], start_string) -> List[str]:
            return [start_string + l for l in lines]

        for i, module in enumerate(submodules):
            lines = module.summary(return_list=True)
            if i + 1 < len(submodules):  # middle submodules
                indented_lines = indent(lines[:1], "├── ") + indent(lines[1:], "│   ")
            else:  # last submodule
                indented_lines = indent(lines[:1], "└── ") + indent(lines[1:], "    ")
            output.extend(indented_lines)

        if return_list:
            return output
        else:
            return "\n".join(output)

    def set_attribute(self: T, name, value) -> T:
        """Create a new module and set attribute."""
        module = self.copy()
        with allow_mutation(module):
            setattr(module, name, value)
            # pylint: disable=protected-access
            module._find_and_register_pytree(PaxKind.MODULE)

        return module



"""Train a transformer language model on TPU (if available)."""

import inspect
import os
from functools import partial
from typing import Tuple

import jax
import jax.numpy as jnp
import jax.tools.colab_tpu
import opax
import pax
from opax import GradientTransformation
from tqdm.auto import tqdm

from data import detokenize, make_data_loader, tokenize
from model import LM


def setup_tpu_device():
    print("Setting up TPU cores")
    jax.tools.colab_tpu.setup_tpu()
    print(jax.devices())


# shared config
dropout = 0.1
learning_rate = 1e-4
vocab_size = 256
pax.seed_rng_key(42)

if "COLAB_TPU_ADDR" in os.environ:
    # TPU config
    # need to config TPU cores _before_ calling `jax.device_count`.
    setup_tpu_device()
    steps_per_update = 50
    num_devices = jax.device_count()
    batch_size = 32 * num_devices * steps_per_update
    seq_len = 256
    hidden_dim = 512
    num_steps = 1_000
    num_layers = 6
else:
    # CPU/GPU config
    steps_per_update = 1
    num_devices = jax.device_count()
    batch_size = 8 * num_devices * steps_per_update
    seq_len = 64
    hidden_dim = 256
    num_steps = 20_000
    num_layers = 2


def loss_fn(model: LM, batch: jnp.ndarray):
    inputs = batch[:, :-1]
    targets = batch[:, 1:]

    model, logits = pax.module_and_value(model)(inputs)
    log_pr = jax.nn.log_softmax(logits, axis=-1)
    targets = jax.nn.one_hot(targets, num_classes=model.vocab_size)
    loss = -jnp.mean(jnp.sum(targets * log_pr, axis=-1))
    return loss, model


def update_step(model_and_optim: Tuple[LM, GradientTransformation], batch: jnp.ndarray):
    model, optimizer = model_and_optim
    (loss, model), grads = pax.value_and_grad(loss_fn, has_aux=True)(model, batch)
    grads = jax.lax.pmean(grads, axis_name="i")
    params = model.parameters()
    optimizer, updates = pax.module_and_value(optimizer)(grads, params)
    params = params.map(jax.lax.sub, updates)
    model = model.update_parameters(params)
    return (model, optimizer), loss


@partial(jax.pmap, axis_name="i")
def update_fn(model: LM, optimizer: GradientTransformation, multi_batch: jnp.ndarray):
    (model, optimizer), losses = pax.scan(update_step, (model, optimizer), multi_batch)
    return model, optimizer, jnp.sum(losses)


def train():
    net = LM(vocab_size=vocab_size, hidden_dim=hidden_dim, num_layers=num_layers)
    print(net.summary())
    optimizer = opax.chain(
        opax.clip_by_global_norm(1.0),
        opax.adam(learning_rate),
    )(net.parameters())

    data = inspect.getsource(LM)  # a _true_ AGI learns about itself.
    test_prompt = data[:20]
    data_iter = make_data_loader(
        data,
        seq_len=seq_len,
        batch_size=batch_size,
        num_devices=num_devices,
        steps_per_update=steps_per_update,
    )

    # replicate on multiple devices
    net = jax.device_put_replicated(net, jax.devices())
    optimizer = jax.device_put_replicated(optimizer, jax.devices())

    total_losses = 0.0
    tr = tqdm(range(0, 1 + num_steps, steps_per_update), desc="training")
    for step in tr:
        batch = next(data_iter)
        # (num_devices,) is for jax.pmap, (steps_per_update,) is for pax.scan
        net, optimizer, loss = update_fn(net, optimizer, batch)
        total_losses = total_losses + loss
        if step % 1000 == 0:
            loss = jnp.mean(total_losses) / (1000 if step > 0 else steps_per_update)
            total_losses = jnp.zeros_like(total_losses)
            # eval on a single device
            eval_net = jax.tree_map(lambda x: x[0], net.eval())
            out = eval_net.inference(
                prompt=tokenize(test_prompt),
                length=(128 if step < num_steps else 1024),
                train_seq_len=seq_len,
            )
            text = detokenize(out)
            tr.write(
                f"[step {step}]  loss {loss:.3f}\n"
                f"Prompt: {test_prompt}\n"
                f"========\n"
                f"{text}\n"
                f"========"
            )


if __name__ == "__main__":
    train()



"""Setup PAX3 package."""

from setuptools import find_namespace_packages, setup


def _get_version():
    with open("pax/__init__.py", encoding="utf-8") as file:
        for line in file:
            if line.startswith("__version__"):
                _globals = {}
                exec(line, _globals)  # pylint: disable=exec-used
                return _globals["__version__"]
        raise ValueError("`__version__` not defined in `pax/__init__.py`")


__version__ = _get_version()
URL = "https://github.com/ntt123/pax"

install_requires = ["jax>=0.2.21", "jmp>=0.0.2"]
setup_requires = []
tests_requires = [
    "chex",
    "dm-haiku",
    "fire",
    "opax",
    "pytest",
    "pytype",
    "tqdm",
]

setup(
    name="pax3",
    version=__version__,
    description="A stateful pytree library for training neural networks.",
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    author="Thông Nguyễn",
    url=URL,
    keywords=[
        "deep-learning",
        "jax",
    ],
    install_requires=install_requires,
    setup_requires=setup_requires,
    tests_require=tests_requires,
    packages=find_namespace_packages(exclude=["examples", "tests"]),
    extras_require={"test": tests_requires},
    python_requires=">=3.7",
    include_package_data=True,
    zip_safe=False,
)
